# copyright 2003-2014 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""Predicate classes
"""

from operator import eq

from logilab.common.registry import Predicate, objectify_predicate
from rql.nodes import Function
from yams.schema import BASE_TYPES, role_name

from cubicweb import (
    Unauthorized,
    NotAnEntity,
    CW_EVENT_MANAGER,
    role,
)
from cubicweb.schema import split_expression
from cubicweb.uilib import eid_param


# abstract predicates / mixin helpers ###########################################


class PartialPredicateMixIn:
    """convenience mix-in for predicates that will look into the containing
    class to find missing information.

    cf. `cubicweb.web.action.LinkToEntityAction` for instance
    """

    def __call__(self, cls, *args, **kwargs):
        self.complete(cls)
        return super(PartialPredicateMixIn, self).__call__(cls, *args, **kwargs)


class EClassPredicate(Predicate):
    """abstract class for predicates working on *entity class(es)* specified
    explicitly or found of the result set.

    Here are entity lookup / scoring rules:

    * if `entity` is specified, return score for this entity's class

    * elif `rset`, `select` and `filtered_variable` are specified, return score
      for the possible classes for variable in the given rql :class:`Select`
      node

    * elif `rset` and `row` are specified, return score for the class of the
      entity found in the specified cell, using column specified by `col` or 0

    * elif `rset` is specified return score for each entity class found in the
      column specified specified by the `col` argument or in column 0 if not
      specified

    When there are several classes to be evaluated, return the sum of scores for
    each entity class unless:

      - `mode` == 'all' (the default) and some entity class is scored
        to 0, in which case 0 is returned

      - `mode` == 'any', in which case the first non-zero score is
        returned

      - `accept_none` is False and some cell in the column has a None value
        (this may occurs with outer join)
    """

    def __init__(self, accept_none=True, mode="all"):
        assert mode in ("any", "all"), f"bad mode {mode}"
        self.once_is_enough = mode == "any"
        self.accept_none = accept_none

    def __call__(
        self,
        cls,
        req,
        rset=None,
        row=None,
        col=0,
        entity=None,
        select=None,
        filtered_variable=None,
        accept_none=None,
        **kwargs,
    ):
        if entity is not None:
            return self.score_class(entity.__class__, req)
        if not rset:
            return 0
        if select is not None and filtered_variable is not None:
            etypes = set(sol[filtered_variable.name] for sol in select.solutions)
        elif row is None:
            if accept_none is None:
                accept_none = self.accept_none
            if not accept_none and any(row[col] is None for row in rset):
                return 0
            etypes = rset.column_types(col)
        else:
            etype = rset.description[row][col]
            # may have None in rset.description on outer join
            if etype is None or rset.rows[row][col] is None:
                return 0
            etypes = (etype,)
        score = 0
        for etype in etypes:
            escore = self.score(cls, req, etype)
            if not escore and not self.once_is_enough:
                return 0
            elif self.once_is_enough:
                return escore
            score += escore
        return score

    def score(self, cls, req, etype):
        if etype in BASE_TYPES:
            return 0
        return self.score_class(req.vreg["etypes"].etype_class(etype), req)

    def score_class(self, eclass, req):
        raise NotImplementedError()


class EntityPredicate(EClassPredicate):
    """abstract class for predicates working on *entity instance(s)* specified
    explicitly or found of the result set.

    Here are entity lookup / scoring rules:

    * if `entity` is specified, return score for this entity

    * elif `row` is specified, return score for the entity found in the
      specified cell, using column specified by `col` or 0

    * else return the sum of scores for each entity found in the column
      specified specified by the `col` argument or in column 0 if not specified,
      unless:

      - `mode` == 'all' (the default) and some entity class is scored
        to 0, in which case 0 is returned

      - `mode` == 'any', in which case the first non-zero score is
        returned

      - `accept_none` is False and some cell in the column has a None value
        (this may occurs with outer join)

    .. Note::
       using :class:`EntityPredicate` or :class:`EClassPredicate` as base predicate
       class impacts performance, since when no entity or row is specified the
       later works on every different *entity class* found in the result set,
       while the former works on each *entity* (eg each row of the result set),
       which may be much more costly.
    """

    def __call__(
        self,
        cls,
        req,
        rset=None,
        row=None,
        col=0,
        accept_none=None,
        entity=None,
        **kwargs,
    ):
        if not rset and entity is None:
            return 0
        score = 0
        if entity is not None:
            score = self.score_entity(entity)
        elif row is None:
            col = col or 0
            if accept_none is None:
                accept_none = self.accept_none
            for row, rowvalue in enumerate(rset.rows):
                if rowvalue[col] is None:  # outer join
                    if not accept_none:
                        return 0
                    continue
                escore = self.score(req, rset, row, col)
                if not escore and not self.once_is_enough:
                    return 0
                elif self.once_is_enough:
                    return escore
                score += escore
        else:
            col = col or 0
            etype = rset.description[row][col]
            if etype is not None:  # outer join
                score = self.score(req, rset, row, col)
        return score

    def score(self, req, rset, row, col):
        try:
            return self.score_entity(rset.get_entity(row, col))
        except NotAnEntity:
            return 0

    def score_entity(self, entity):
        raise NotImplementedError()


class ExpectedValuePredicate(Predicate):
    """Take a list of expected values as initializer argument and store them
    into the :attr:`expected` set attribute. You may also give a set as single
    argument, which will then be referenced as set of expected values,
    allowing modifications to the given set to be considered.

    You should implement one of :meth:`_values_set(cls, req, **kwargs)` or
    :meth:`_get_value(cls, req, **kwargs)` method which should respectively
    return the set of values or the unique possible value for the given context.

    You may also specify a `mode` behaviour as argument, as explained below.

    Returned score is:

    - 0 if `mode` == 'all' (the default) and at least one expected
      values isn't found

    - 0 if `mode` == 'any' and no expected values isn't found at all

    - else the number of matching values

    Notice `mode` = 'any' with a single expected value has no effect at all.
    """

    def __init__(self, *expected, **kwargs):
        assert expected, self
        if len(expected) == 1 and isinstance(expected[0], (set, dict)):
            self.expected = expected[0]
        else:
            self.expected = frozenset(expected)
        mode = kwargs.pop("mode", "all")
        assert mode in ("any", "all"), f"bad mode {mode}"
        self.once_is_enough = mode == "any"
        assert not kwargs, f"unexpected arguments {kwargs}"

    def __str__(self):
        return "%s(%s)" % (
            self.__class__.__name__,
            ",".join(sorted(str(s) for s in self.expected)),
        )

    def __call__(self, cls, req, **kwargs):
        values = self._values_set(cls, req, **kwargs)
        if isinstance(values, dict):
            if isinstance(self.expected, dict):
                matching = 0
                for key, expected_value in self.expected.items():
                    if key in values:
                        if (
                            isinstance(expected_value, (list, tuple, frozenset, set))
                            and values[key] in expected_value
                        ):
                            matching += 1
                        elif values[key] == expected_value:
                            matching += 1
            if isinstance(self.expected, (set, frozenset)):
                values = frozenset(values)
                matching = len(values & self.expected)
        else:
            matching = len(values & self.expected)
        if self.once_is_enough:
            return matching
        if matching == len(self.expected):
            return matching
        return 0

    def _values_set(self, cls, req, **kwargs):
        return frozenset((self._get_value(cls, req, **kwargs),))

    def _get_value(self, cls, req, **kwargs):
        raise NotImplementedError()


# bare predicates ##############################################################


class match_kwargs(ExpectedValuePredicate):
    """Return non-zero score if parameter names specified as initializer
    arguments are specified in the input context.


    Return a score corresponding to the number of expected parameters.

    When multiple parameters are expected, all of them should be found in
    the input context unless `mode` keyword argument is given to 'any',
    in which case a single matching parameter is enough.
    """

    def _values_set(self, cls, req, **kwargs):
        return kwargs


class appobject_selectable(Predicate):
    """Return 1 if another appobject is selectable using the same input context.

    Initializer arguments:

    * `registry`, a registry name

    * `regids`, object identifiers in this registry, one of them should be
      selectable.
    """

    selectable_score = 1

    def __init__(self, registry, *regids):
        self.registry = registry
        self.regids = regids

    def __call__(self, cls, req, **kwargs):
        for regid in self.regids:
            if req.vreg[self.registry].select_or_none(regid, req, **kwargs) is not None:
                return self.selectable_score
        return 0


class adaptable(appobject_selectable):
    """Return 1 if another appobject is selectable using the same input context.

    Initializer arguments:

    * `regids`, adapter identifiers (e.g. interface names) to which the context
      (usually entities) should be adaptable. One of them should be selectable
      when multiple identifiers are given.
    """

    def __init__(self, *regids):
        super(adaptable, self).__init__("adapters", *regids)

    def __call__(self, cls, req, **kwargs):
        kwargs.setdefault("accept_none", False)
        score = super(adaptable, self).__call__(cls, req, **kwargs)
        if (
            score == 0
            and kwargs.get("rset")
            and len(kwargs["rset"]) > 1
            and "row" not in kwargs
        ):
            # on rset containing several entity types, each row may be
            # individually adaptable, while the whole rset won't be if the
            # same adapter can't be used for each type
            for row in range(len(kwargs["rset"])):
                kwargs.setdefault("col", 0)
                _score = super(adaptable, self).__call__(cls, req, row=row, **kwargs)
                if not _score:
                    return 0
                # adjust score per row as expected by default adjust_score
                # implementation
                score += self.adjust_score(_score)
        else:
            score = self.adjust_score(score)
        return score

    @staticmethod
    def adjust_score(score):
        # being adaptable to an interface should takes precedence other
        # is_instance('Any'), but not other explicit
        # is_instance('SomeEntityType'), and, for **a single entity**:
        # * is_instance('Any') score is 1
        # * is_instance('SomeEntityType') score is at least 2
        if score >= 2:
            return score - 0.5
        if score == 1:
            return score + 0.5
        return score


class configuration_values(Predicate):
    """Return 1 if the instance has an option set to a given value(s) in its
    configuration file.
    """

    # XXX this predicate could be evaluated on startup
    def __init__(self, key, values):
        self._key = key
        if not isinstance(values, (tuple, list)):
            values = (values,)
        self._values = frozenset(values)

    def __call__(self, cls, req, **kwargs):
        if hasattr(self, "_score"):
            return self._score

        if req is None:
            config = kwargs["repo"].config
        else:
            config = req.vreg.config

        self._score = config[self._key] in self._values

        return self._score


# rset predicates ##############################################################


@objectify_predicate
def none_rset(cls, req, rset=None, **kwargs):
    """Return 1 if the result set is None (eg usually not specified)."""
    if rset is None:
        return 1
    return 0


# XXX == ~ none_rset
@objectify_predicate
def any_rset(cls, req, rset=None, **kwargs):
    """Return 1 for any result set, whatever the number of rows in it, even 0."""
    if rset is not None:
        return 1
    return 0


@objectify_predicate
def nonempty_rset(cls, req, rset=None, **kwargs):
    """Return 1 for result set containing one ore more rows."""
    if rset:
        return 1
    return 0


# XXX == ~ nonempty_rset
@objectify_predicate
def empty_rset(cls, req, rset=None, **kwargs):
    """Return 1 for result set which doesn't contain any row."""
    if rset is not None and len(rset) == 0:
        return 1
    return 0


# XXX == multi_lines_rset(1)
@objectify_predicate
def one_line_rset(cls, req, rset=None, row=None, **kwargs):
    """Return 1 if the result set is of size 1, or greater but a specific row in
    the result set is specified ('row' argument).
    """
    if rset is None and "entity" in kwargs:
        return 1
    if rset is not None and (row is not None or len(rset) == 1):
        return 1
    return 0


class multi_lines_rset(Predicate):
    """Return 1 if the operator expression matches between `num` elements
    in the result set and the `expected` value if defined.

    By default, multi_lines_rset(expected) matches equality expression:
        `nb` row(s) in result set equals to expected value
    But, you can perform richer comparisons by overriding default operator:
        multi_lines_rset(expected, operator.gt)

    If `expected` is None, return 1 if the result set contains *at least*
    two rows.
    If rset is None, return 0.
    """

    def __init__(self, expected=None, operator=eq):
        self.expected = expected
        self.operator = operator

    def match_expected(self, num):
        if self.expected is None:
            return num > 1
        return self.operator(num, self.expected)

    def __call__(self, cls, req, rset=None, **kwargs):
        return int(rset is not None and self.match_expected(len(rset)))


class multi_columns_rset(multi_lines_rset):
    """If `nb` is specified, return 1 if the result set has exactly `nb` column
    per row. Else (`nb` is None), return 1 if the result set contains *at least*
    two columns per row. Return 0 for empty result set.
    """

    def __call__(self, cls, req, rset=None, **kwargs):
        # 'or 0' since we *must not* return None. Also don't use rset.rows so
        # this selector will work if rset is a simple list of list.
        return rset and self.match_expected(len(rset[0])) or 0


@objectify_predicate
def sorted_rset(cls, req, rset=None, **kwargs):
    """Return 1 for sorted result set (e.g. from an RQL query containing an
    ORDERBY clause), with exception that it will return 0 if the rset is
    'ORDERBY FTIRANK(VAR)' (eg sorted by rank value of the has_text index).
    """
    if rset is None:
        return 0
    selects = rset.syntax_tree().children
    if (
        len(selects) > 1
        or not selects[0].orderby
        or (
            isinstance(selects[0].orderby[0].term, Function)
            and selects[0].orderby[0].term.name == "FTIRANK"
        )
    ):
        return 0
    return 2


# XXX == multi_etypes_rset(1)
@objectify_predicate
def one_etype_rset(cls, req, rset=None, col=0, **kwargs):
    """Return 1 if the result set contains entities which are all of the same
    type in the column specified by the `col` argument of the input context, or
    in column 0.
    """
    if rset is None:
        return 0
    if len(rset.column_types(col)) != 1:
        return 0
    return 1


class multi_etypes_rset(multi_lines_rset):
    """If `nb` is specified, return 1 if the result set contains `nb` different
    types of entities in the column specified by the `col` argument of the input
    context, or in column 0. If `nb` is None, return 1 if the result set contains
    *at least* two different types of entities.
    """

    def __call__(self, cls, req, rset=None, col=0, **kwargs):
        # 'or 0' since we *must not* return None
        return rset and self.match_expected(len(rset.column_types(col))) or 0


@objectify_predicate
def logged_user_in_rset(cls, req, rset=None, row=None, col=0, **kwargs):
    """Return positive score if the result set at the specified row / col
    contains the eid of the logged user.
    """
    if rset is None:
        return 0
    return req.user.eid == rset[row or 0][col]


# entity predicates #############################################################


class composite_etype(Predicate):
    """Return 1 for composite entities.

    A composite entity has an etype for which at least one relation
    definition points in its direction with the
    composite='subject'/'object' notation.
    """

    def __call__(self, cls, req, **kwargs):
        entity = kwargs.pop("entity", None)
        if entity is None:
            return 0
        return entity.e_schema.is_composite


class non_final_entity(EClassPredicate):
    """Return 1 for entity of a non final entity type(s). Remember, "final"
    entity types are String, Int, etc... This is equivalent to
    `is_instance('Any')` but more optimized.

    See :class:`~cubicweb.predicates.EClassPredicate` documentation for entity
    class lookup / score rules according to the input context.
    """

    def score(self, cls, req, etype):
        if etype in BASE_TYPES:
            return 0
        return 1

    def score_class(self, eclass, req):
        return 1  # necessarily true if we're there


def _reset_is_instance_cache(vreg):
    vreg._is_instance_predicate_cache = {}


CW_EVENT_MANAGER.bind("before-registry-reset", _reset_is_instance_cache)


class is_instance(EClassPredicate):
    """Return non-zero score for entity that is an instance of the one of given
    type(s). If multiple arguments are given, matching one of them is enough.

    Entity types should be given as string, the corresponding class will be
    fetched from the registry at selection time.

    See :class:`~cubicweb.predicates.EClassPredicate` documentation for entity
    class lookup / score rules according to the input context.

    .. note:: the score will reflect class proximity so the most specific object
              will be selected.
    """

    def __init__(self, *expected_etypes, **kwargs):
        super(is_instance, self).__init__(**kwargs)
        self.expected_etypes = expected_etypes
        for etype in self.expected_etypes:
            assert isinstance(etype, str), etype

    def __str__(self):
        return "%s(%s)" % (
            self.__class__.__name__,
            ",".join(str(s) for s in self.expected_etypes),
        )

    def score_class(self, eclass, req):
        if not hasattr(req.vreg, "_is_instance_predicate_cache"):
            # XXX 'before-registry-reset' not called for db-api connections
            req.vreg._is_instance_predicate_cache = {}

        # cache on vreg to avoid reloading issues
        cache = req.vreg._is_instance_predicate_cache

        if self in cache:
            expected_eclasses = cache[self]
        else:
            # turn list of entity types as string into a list of
            #  (entity class, parent classes)
            etypesreg = req.vreg["etypes"]
            expected_eclasses = cache[self] = []
            for etype in self.expected_etypes:
                try:
                    expected_eclasses.append(etypesreg.etype_class(etype))
                except KeyError:
                    continue  # entity type not in the schema

        parents, any = req.vreg["etypes"].parent_classes(eclass.__regid__)
        score = 0
        for expectedcls in expected_eclasses:
            # adjust score according to class proximity
            if expectedcls is eclass:
                score += len(parents) + 4
            elif expectedcls is any:  # Any
                score += 1
            else:
                for index, basecls in enumerate(reversed(parents)):
                    if expectedcls is basecls:
                        score += index + 3
                        break
        return score


class score_entity(EntityPredicate):
    """Return score according to an arbitrary function given as argument which
    will be called with input content entity as argument.

    This is a very useful predicate that will usually interest you since it
    allows a lot of things without having to write a specific predicate.

    The function can return arbitrary value which will be casted to an integer
    value at the end.

    See :class:`~cubicweb.predicates.EntityPredicate` documentation for entity
    lookup / score rules according to the input context.
    """

    def __init__(self, scorefunc, mode="all"):
        super(score_entity, self).__init__(mode=mode)

        def intscore(*args, **kwargs):
            score = scorefunc(*args, **kwargs)
            if not score:
                return 0
            if isinstance(score, int):
                return score
            return 1

        self.score_entity = intscore


class has_mimetype(EntityPredicate):
    """Return 1 if the entity adapt to IDownloadable and has the given MIME type.

    You can give 'image/' to match any image for instance, or 'image/png' to match
    only PNG images.
    """

    def __init__(self, mimetype, mode="all"):
        super(has_mimetype, self).__init__(mode=mode)
        self.mimetype = mimetype

    def score_entity(self, entity):
        idownloadable = entity.cw_adapt_to("IDownloadable")
        if idownloadable is None:
            return 0
        mt = idownloadable.download_content_type()
        if not (mt and mt.startswith(self.mimetype)):
            return 0
        return 1


class relation_possible(EntityPredicate):
    """Return 1 for entity that supports the relation, provided that the
    request's user may do some `action` on it (see below).

    The relation is specified by the following initializer arguments:

    * `rtype`, the name of the relation

    * `role`, the role of the entity in the relation, either 'subject' or
      'object', default to 'subject'

    * `target_etype`, optional name of an entity type that should be supported
      at the other end of the relation

    * `action`, a relation schema action (e.g. one of 'read', 'add', 'delete',
      default to 'read') which must be granted to the user, else a 0 score will
      be returned. Give None if you don't want any permission checking.

    * `strict`, boolean (default to False) telling what to do when the user has
      not globally the permission for the action (eg the action is not granted
      to one of the user's groups)

      - when strict is False, if there are some local role defined for this
        action (e.g. using rql expressions), then the permission will be
        considered as granted

      - when strict is True, then the permission will be actually checked for
        each entity

    Setting `strict` to True impacts performance for large result set since
    you'll then get the :class:`~cubicweb.predicates.EntityPredicate` behaviour
    while otherwise you get the :class:`~cubicweb.predicates.EClassPredicate`'s
    one. See those classes documentation for entity lookup / score rules
    according to the input context.
    """

    def __init__(
        self,
        rtype,
        role="subject",
        target_etype=None,
        action="read",
        strict=False,
        **kwargs,
    ):
        super(relation_possible, self).__init__(**kwargs)
        self.rtype = rtype
        self.role = role
        self.target_etype = target_etype
        self.action = action
        self.strict = strict

    # hack hack hack
    def __call__(self, cls, req, **kwargs):
        # hack hack hack
        if self.strict:
            return EntityPredicate.__call__(self, cls, req, **kwargs)
        return EClassPredicate.__call__(self, cls, req, **kwargs)

    def score(self, *args):
        if self.strict:
            return EntityPredicate.score(self, *args)
        return EClassPredicate.score(self, *args)

    def _get_rschema(self, eclass):
        eschema = eclass.e_schema

        if self.role == "object":
            return eschema.object_relations.get(self.rtype)
        else:
            return eschema.subject_relations.get(self.rtype)

    def score_class(self, eclass, req):
        rschema = self._get_rschema(eclass)
        if rschema is None:
            return 0  # relation not supported

        eschema = eclass.e_schema
        if self.target_etype is not None:
            try:
                rdef = rschema.role_relation_definition(
                    eschema, self.target_etype, self.role
                )
            except KeyError:
                return 0

            if self.action and not rdef.may_have_permission(self.action, req):
                return 0

            teschema = req.vreg.schema.entity_schema_for(self.target_etype)
            if not teschema.may_have_permission("read", req):
                return 0

        elif self.action:
            return rschema.may_have_permission(self.action, req, eschema, self.role)

        return 1

    def score_entity(self, entity):
        rschema = self._get_rschema(entity)
        if rschema is None:
            return 0  # relation not supported

        if self.action:
            if self.target_etype is not None:
                try:
                    rschema = rschema.role_relation_definition(
                        entity.e_schema, self.target_etype, self.role
                    )
                except KeyError:
                    return 0

            if self.role == "subject":
                if not rschema.has_perm(entity._cw, self.action, fromeid=entity.eid):
                    return 0

            elif not rschema.has_perm(entity._cw, self.action, toeid=entity.eid):
                return 0

        if self.target_etype is not None:
            req = entity._cw
            teschema = req.vreg.schema.entity_schema_for(self.target_etype)

            if not teschema.may_have_permission("read", req):
                return 0
        return 1


class partial_relation_possible(PartialPredicateMixIn, relation_possible):
    """Same as :class:~`cubicweb.predicates.relation_possible`, but will look for
    attributes of the selected class to get information which is otherwise
    expected by the initializer, except for `action` and `strict` which are kept
    as initializer arguments.

    This is useful to predefine predicate of an abstract class designed to be
    customized.
    """

    def __init__(self, action="read", **kwargs):
        super(partial_relation_possible, self).__init__(
            None, None, None, action, **kwargs
        )

    def complete(self, cls):
        self.rtype = cls.rtype
        self.role = role(cls)
        self.target_etype = getattr(cls, "target_etype", None)


class has_related_entities(EntityPredicate):
    """Return 1 if entity support the specified relation and has some linked
    entities by this relation , optionally filtered according to the specified
    target type.

    The relation is specified by the following initializer arguments:

    * `rtype`, the name of the relation

    * `role`, the role of the entity in the relation, either 'subject' or
      'object', default to 'subject'.

    * `target_etype`, optional name of an entity type that should be found
      at the other end of the relation

    See :class:`~cubicweb.predicates.EntityPredicate` documentation for entity
    lookup / score rules according to the input context.
    """

    def __init__(self, rtype, role="subject", target_etype=None, **kwargs):
        super(has_related_entities, self).__init__(**kwargs)
        self.rtype = rtype
        self.role = role
        self.target_etype = target_etype

    def score_entity(self, entity):
        relpossel = relation_possible(self.rtype, self.role, self.target_etype)
        if not relpossel.score_class(entity.__class__, entity._cw):
            return 0
        rset = entity.related(self.rtype, self.role)
        if self.target_etype:
            return any(r for r in rset.description if r[0] == self.target_etype)
        return rset and 1 or 0


class partial_has_related_entities(PartialPredicateMixIn, has_related_entities):
    """Same as :class:~`cubicweb.predicates.has_related_entities`, but will look
    for attributes of the selected class to get information which is otherwise
    expected by the initializer.

    This is useful to predefine predicate of an abstract class designed to be
    customized.
    """

    def __init__(self, **kwargs):
        super(partial_has_related_entities, self).__init__(None, None, None, **kwargs)

    def complete(self, cls):
        self.rtype = cls.rtype
        self.role = role(cls)
        self.target_etype = getattr(cls, "target_etype", None)


class has_permission(EntityPredicate):
    """Return non-zero score if request's user has the permission to do the
    requested action on the entity. `action` is an entity schema action (eg one
    of 'read', 'add', 'delete', 'update').

    Here are entity lookup / scoring rules:

    * if `entity` is specified, check permission is granted for this entity

    * elif `row` is specified, check permission is granted for the entity found
      in the specified cell

    * else check permission is granted for each entity found in the column
      specified specified by the `col` argument or in column 0
    """

    def __init__(self, action):
        self.action = action

    # don't use EntityPredicate.__call__ but this optimized implementation to
    # avoid considering each entity when it's not necessary
    def __call__(self, cls, req, rset=None, row=None, col=0, entity=None, **kwargs):
        if entity is not None:
            return self.score_entity(entity)
        if rset is None:
            return 0
        if row is None:
            score = 0
            need_local_check = []
            geteschema = req.vreg.schema.entity_schema_for
            user = req.user
            action = self.action
            for etype in rset.column_types(0):
                if etype in BASE_TYPES:
                    return 0
                eschema = geteschema(etype)
                if not user.matching_groups(eschema.get_groups(action)):
                    if eschema.has_local_role(action):
                        # have to ckeck local roles
                        need_local_check.append(eschema)
                        continue
                    else:
                        # even a local role won't be enough
                        return 0
                score += 1
            if need_local_check:
                # check local role for entities of necessary types
                for i, row in enumerate(rset):
                    if rset.description[i][col] not in need_local_check:
                        continue
                    # micro-optimisation instead of calling self.score(req,
                    # rset, i, col): rset may be large
                    if not rset.get_entity(i, col).cw_has_perm(action):
                        return 0
                score += 1
            return score
        return self.score(req, rset, row, col)

    def score_entity(self, entity):
        if entity.cw_has_perm(self.action):
            return 1
        return 0


class has_add_permission(EClassPredicate):
    """Return 1 if request's user has the add permission on entity type
    specified in the `etype` initializer argument, or according to entity found
    in the input content if not specified.

    It also check that then entity type is not a strict subobject (e.g. may only
    be used as a composed of another entity).

    See :class:`~cubicweb.predicates.EClassPredicate` documentation for entity
    class lookup / score rules according to the input context when `etype` is
    not specified.
    """

    def __init__(self, etype=None, **kwargs):
        super(has_add_permission, self).__init__(**kwargs)
        self.etype = etype

    def __call__(self, cls, req, **kwargs):
        if self.etype is None:
            return super(has_add_permission, self).__call__(cls, req, **kwargs)
        return self.score(cls, req, self.etype)

    def score_class(self, eclass, req):
        eschema = eclass.e_schema
        if (
            eschema.final
            or eschema.is_subobject(strict=True)
            or not eschema.has_perm(req, "add")
        ):
            return 0
        return 1


class rql_condition(EntityPredicate):
    """Return non-zero score if arbitrary rql specified in `expression`
    initializer argument return some results for entity found in the input
    context. Returned score is the number of items returned by the rql
    condition.

    `expression` is expected to be a string containing an rql expression, which
    must use 'X' variable to represent the context entity and may use 'U' to
    represent the request's user.

    .. warning::
        If simply testing value of some attribute/relation of context entity (X),
        you should rather use the :class:`score_entity` predicate which will
        benefit from the ORM's request entities cache.

    See :class:`~cubicweb.predicates.EntityPredicate` documentation for entity
    lookup / score rules according to the input context.
    """

    def __init__(self, expression, mode="all", user_condition=False):
        super(rql_condition, self).__init__(mode=mode)
        self.user_condition = user_condition
        if user_condition:
            rql = "Any COUNT(U) WHERE U eid %%(u)s, %s" % expression
        elif "U" in frozenset(split_expression(expression)):
            rql = "Any COUNT(X) WHERE X eid %%(x)s, U eid %%(u)s, %s" % expression
        else:
            rql = "Any COUNT(X) WHERE X eid %%(x)s, %s" % expression
        self.rql = rql

    def __str__(self):
        return f"{self.__class__.__name__}({self.rql!r})"

    def __call__(self, cls, req, **kwargs):
        if self.user_condition:
            try:
                return req.execute(self.rql, {"u": req.user.eid})[0][0]
            except Unauthorized:
                return 0
        else:
            return super(rql_condition, self).__call__(cls, req, **kwargs)

    def _score(self, req, eid):
        try:
            return req.execute(self.rql, {"x": eid, "u": req.user.eid})[0][0]
        except Unauthorized:
            return 0

    def score(self, req, rset, row, col):
        return self._score(req, rset[row][col])

    def score_entity(self, entity):
        return self._score(entity._cw, entity.eid)


# workflow predicates ###########################################################


class is_in_state(score_entity):
    """Return 1 if entity is in one of the states given as argument list

    You should use this instead of your own :class:`score_entity` predicate to
    avoid some gotchas:

    * possible views gives a fake entity with no state
    * you must use the latest tr info thru the workflow adapter for repository
      side checking of the current state

    In debug mode, this predicate can raise :exc:`ValueError` for unknown states names
    (only checked on entities without a custom workflow)

    :rtype: int
    """

    def __init__(self, *expected):
        assert expected, self
        self.expected = frozenset(expected)

        def score(entity, expected=self.expected):
            adapted = entity.cw_adapt_to("IWorkflowable")
            # in debug mode only (time consuming)
            if entity._cw.vreg.config.debugmode:
                # validation can only be done for generic etype workflow because
                # expected transition list could have been changed for a custom
                # workflow (for the current entity)
                if not entity.custom_workflow:
                    self._validate(adapted)
            return self._score(adapted)

        super(is_in_state, self).__init__(score)

    def _score(self, adapted):
        trinfo = adapted.latest_trinfo()
        if trinfo is None:  # entity is probably in it's initial state
            statename = adapted.state
        else:
            statename = trinfo.new_state.name
        return statename in self.expected

    def _validate(self, adapted):
        wf = adapted.current_workflow
        valid = [n.name for n in wf.reverse_state_of]
        unknown = sorted(self.expected.difference(valid))
        if unknown:
            raise ValueError(f"{wf.name}: unknown state(s): {','.join(unknown)}")

    def __str__(self):
        return f"{self.__class__.__name__}({','.join(str(s) for s in self.expected)})"


def on_fire_transition(etype, tr_names):
    """Return 1 when entity of the type `etype` is going through transition of
    a name included in `tr_names`.

    You should use this predicate on 'after_add_entity' hook, since it's actually
    looking for addition of `TrInfo` entities. Hence in the hook, `self.entity`
    will reference the matching `TrInfo` entity, allowing to get all the
    transition details (including the entity to which is applied the transition
    but also its original state, transition, destination state, user...).

    See :class:`cubicweb.entities.wfobjs.TrInfo` for more information.
    """
    if isinstance(tr_names, str):
        tr_names = set((tr_names,))

    def match_etype_and_transition(trinfo):
        # take care trinfo.transition is None when calling change_state
        return (
            trinfo.transition
            and trinfo.transition.name in tr_names
            # is_instance() first two arguments are 'cls' (unused, so giving
            # None is fine) and the request/session
            and is_instance(etype)(None, trinfo._cw, entity=trinfo.for_entity)
        )

    return is_instance("TrInfo") & score_entity(match_etype_and_transition)


class match_transition(ExpectedValuePredicate):
    """Return 1 if `transition` argument is found in the input context which has
    a `.name` attribute matching one of the expected names given to the
    initializer.

    This predicate is expected to be used to customise the status change form in
    the web ui.
    """

    def __call__(self, cls, req, transition=None, **kwargs):
        # XXX check this is a transition that apply to the object?
        if transition is None:
            treid = req.form.get("treid", None)
            if treid:
                transition = req.entity_from_eid(treid)
        if (
            transition is not None
            and getattr(transition, "name", None) in self.expected
        ):
            return 1
        return 0


# logged user predicates ########################################################


@objectify_predicate
def no_cnx(cls, req, **kwargs):
    """Return 1 if the web session has no connection set. This occurs when
    anonymous access is not allowed and user isn't authenticated.
    """
    if not req.cnx:
        return 1
    return 0


@objectify_predicate
def authenticated_user(cls, req, **kwargs):
    """Return 1 if the user is authenticated (i.e. not the anonymous user)."""
    if req.session.anonymous_session:
        return 0
    return 1


@objectify_predicate
def anonymous_user(cls, req, **kwargs):
    """Return 1 if the user is not authenticated (i.e. is the anonymous user)."""
    if req.session.anonymous_session:
        return 1
    return 0


class match_user_groups(ExpectedValuePredicate):
    """Return a non-zero score if request's user is in at least one of the
    groups given as initializer argument. Returned score is the number of groups
    in which the user is.

    If the special 'owners' group is given and `rset` is specified in the input
    context:

    * if `row` is specified check the entity at the given `row`/`col` (default
      to 0) is owned by the user

    * else check all entities in `col` (default to 0) are owned by the user
    """

    def __call__(self, cls, req, rset=None, row=None, col=0, **kwargs):
        if not getattr(req, "cnx", True):  # default to True for repo session instances
            return 0
        user = req.user
        if user is None:
            return int("guests" in self.expected)
        score = user.matching_groups(self.expected)
        if not score and "owners" in self.expected and rset:
            if row is not None:
                if not user.owns(rset[row][col]):
                    return 0
                score = 1
            else:
                score = all(user.owns(r[col]) for r in rset)
        return score


# Web request predicates ########################################################


# XXX deprecate
@objectify_predicate
def primary_view(cls, req, view=None, **kwargs):
    """Return 1 if:

    * *no view is specified* in the input context

    * a view is specified and its `.is_primary()` method return True

    This predicate is usually used by contextual components that only want to
    appears for the primary view of an entity.
    """
    if view is not None and not view.is_primary():
        return 0
    return 1


@objectify_predicate
def contextual(cls, req, view=None, **kwargs):
    """Return 1 if view's contextual property is true"""
    if view is not None and view.contextual:
        return 1
    return 0


class match_view(ExpectedValuePredicate):
    """Return 1 if a view is specified an as its registry id is in one of the
    expected view id given to the initializer.
    """

    def __call__(self, cls, req, view=None, **kwargs):
        if view is None or view.__regid__ not in self.expected:
            return 0
        return 1


class match_context(ExpectedValuePredicate):
    def __call__(self, cls, req, context=None, **kwargs):
        if context not in self.expected:
            return 0
        return 1


# XXX deprecate
@objectify_predicate
def match_context_prop(cls, req, context=None, **kwargs):
    """Return 1 if:

    * no `context` is specified in input context (take care to confusion, here
      `context` refers to a string given as an argument to the input context...)

    * specified `context` is matching the context property value for the
      appobject using this predicate

    * the appobject's context property value is None

    This predicate is usually used by contextual components that want to appears
    in a configurable place.
    """
    if context is None:
        return 1
    propval = req.property_value(f"{cls.__registry__}.{cls.__regid__}.context")
    if propval and context != propval:
        return 0
    return 1


class match_search_state(ExpectedValuePredicate):
    """Return 1 if the current request search state is in one of the expected
    states given to the initializer.

    Known search states are either 'normal' or 'linksearch' (eg searching for an
    object to create a relation with another).

    This predicate is usually used by action that want to appears or not according
    to the ui search state.
    """

    def __call__(self, cls, req, **kwargs):
        try:
            if req.search_state[0] not in self.expected:
                return 0
        except AttributeError:
            return 1  # class doesn't care about search state, accept it
        return 1


class match_form_params(ExpectedValuePredicate):
    """Return non-zero score if parameter names specified as initializer
    arguments are specified in request's form parameters.

    Return a score corresponding to the number of expected parameters.

    When multiple parameters are expected, all of them should be found in
    the input context unless `mode` keyword argument is given to 'any',
    in which case a single matching parameter is enough.
    """

    def __init__(self, *expected, **kwargs):
        """override default __init__ to allow either named or positional
        parameters.
        """
        if kwargs and expected:
            raise ValueError(
                "match_form_params() can't be called with both "
                "positional and named arguments"
            )
        if expected:
            if len(expected) == 1 and not isinstance(expected[0], str):
                raise ValueError(
                    "match_form_params() positional arguments " "must be strings"
                )
            super(match_form_params, self).__init__(*expected)
        else:
            super(match_form_params, self).__init__(kwargs)

    def _values_set(self, cls, req, **kwargs):
        return req.form


class match_http_method(ExpectedValuePredicate):
    """Return non-zero score if one of the HTTP methods specified as
    initializer arguments is the HTTP method of the request (GET, POST, ...).
    """

    def __call__(self, cls, req, **kwargs):
        return int(req.http_method() in self.expected)


class match_edited_type(ExpectedValuePredicate):
    """return non-zero if main edited entity type is the one specified as
    initializer argument, or is among initializer arguments if `mode` == 'any'.
    """

    def _values_set(self, cls, req, **kwargs):
        try:
            return frozenset((req.form[f"__type:{req.form['__maineid']}"],))
        except KeyError:
            return frozenset()


class match_form_id(ExpectedValuePredicate):
    """return non-zero if request form identifier is the one specified as
    initializer argument, or is among initializer arguments if `mode` == 'any'.
    """

    def _values_set(self, cls, req, **kwargs):
        if "__form_id" not in req.form:
            return frozenset()

        return frozenset((req.form["__form_id"],))


class specified_etype_implements(is_instance):
    """Return non-zero score if the entity type specified by an 'etype' key
    searched in (by priority) input context kwargs and request form parameters
    match a known entity type (case insensitivly), and it's associated entity
    class is of one of the type(s) given to the initializer. If multiple
    arguments are given, matching one of them is enough.

    .. note:: as with :class:`~cubicweb.predicates.is_instance`, entity types
              should be given as string and the score will reflect class
              proximity so the most specific object will be selected.

    This predicate is usually used by views holding entity creation forms (since
    we've no result set to work on).
    """

    def __call__(self, cls, req, **kwargs):
        if "etype" in kwargs:
            etype = kwargs["etype"]
        else:
            if "etype" not in req.form:
                return 0

            etype = req.form["etype"]

            if etype.lower() not in req.vreg.case_insensitive_etypes:
                return 0

            # only check this is a known type if etype comes from req.form,
            # else we want the error to propagate
            etype = req.vreg.case_insensitive_etypes[etype.lower()]
            req.form["etype"] = etype

        score = self.score_class(req.vreg["etypes"].etype_class(etype), req)
        if score:
            eschema = req.vreg.schema.entity_schema_for(etype)
            if eschema.may_have_permission("add", req):
                return score

        return 0


class attribute_edited(EntityPredicate):
    """Scores if the specified attribute has been edited This is useful for
    selection of forms by the edit controller.

    The initial use case is on a form, in conjunction with match_transition,
    which will not score at edit time::

     is_instance('Version') & (match_transition('ready') |
                               attribute_edited('publication_date'))
    """

    def __init__(self, attribute, mode="all"):
        super(attribute_edited, self).__init__(mode=mode)
        self._attribute = attribute

    def score_entity(self, entity):
        return (
            eid_param(role_name(self._attribute, "subject"), entity.eid)
            in entity._cw.form
        )


# Other predicates ##############################################################


class match_exception(ExpectedValuePredicate):
    """Return 1 if exception given as `exc` in the input context is an instance
    of one of the class given on instanciation of this predicate.
    """

    def __init__(self, *expected):
        assert expected, self
        # we want a tuple, not a set as done in the parent class
        self.expected = expected

    def __call__(self, cls, req, exc=None, **kwargs):
        if exc is not None and isinstance(exc, self.expected):
            return 1
        return 0


@objectify_predicate
def debug_mode(cls, req, rset=None, **kwargs):
    """Return 1 if running in debug mode."""
    return req.vreg.config.debugmode and 1 or 0



# copyright 2003-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""Define server side service provided by cubicweb"""

import threading

from cubicweb.predicates import match_user_groups, match_kwargs
from cubicweb.server import Service


class StatsService(Service):
    """Return a dictionary containing some statistics about the repository
    resources usage.
    """

    __regid__ = "repo_stats"
    __select__ = match_user_groups("managers", "users")

    def call(self):
        repo = self._cw.repo  # Service are repo side only.
        results = {}
        querier = repo.querier
        source = repo.system_source
        for size, maxsize, hits, misses, title in (
            (
                len(querier.rql_cache),
                repo.config["rql-cache-size"],
                querier.rql_cache.cache_hit,
                querier.rql_cache.cache_miss,
                "rqlt_st",
            ),
            (
                len(source._cache),
                repo.config["rql-cache-size"],
                source.cache_hit,
                source.cache_miss,
                "sql",
            ),
        ):
            results[f"{title}_cache_size"] = {"size": size, "maxsize": maxsize}
            results[f"{title}_cache_hit"] = hits
            results[f"{title}_cache_miss"] = misses
            results[f"{title}_cache_hit_percent"] = (hits * 100) / (hits + misses)
        results["type_cache_size"] = len(repo._type_cache)
        results["sql_no_cache"] = repo.system_source.no_cache
        results["nb_active_threads"] = threading.active_count()
        results["available_cnxsets"] = repo.cnxsets.qsize()
        results["threads"] = [t.name for t in threading.enumerate()]
        return results


class GcStatsService(Service):
    """Return a dictionary containing some statistics about the repository
    resources usage.
    """

    __regid__ = "repo_gc_stats"
    __select__ = match_user_groups("managers")

    def call(self, nmax=20):
        """Return a dictionary containing some statistics about the repository
        memory usage.

        nmax is the max number of (most) referenced object returned as
        the 'referenced' result
        """

        from cubicweb._gcdebug import gc_info
        from cubicweb.appobject import AppObject
        from cubicweb.rset import ResultSet
        from rql.stmts import Union

        lookupclasses = (AppObject, Union, ResultSet)

        results = {}
        counters, ocounters, garbage = gc_info(lookupclasses, viewreferrersclasses=())
        values = sorted(counters.items(), key=lambda x: x[1], reverse=True)
        results["lookupclasses"] = values
        values = sorted(ocounters.items(), key=lambda x: x[1], reverse=True)[:nmax]
        results["referenced"] = values
        results["unreachable"] = garbage
        return results


class RegisterUserService(Service):
    """check if a user with the given login exists, if not create it with the
    given password. This service is designed to be used for anonymous
    registration on public web sites.

    To use it, do:
     with self.appli.repo.internal_cnx() as cnx:
        cnx.call_service('register_user',
                         login=login,
                         password=password,
                         **cwuserkwargs)
    """

    __regid__ = "register_user"
    __select__ = Service.__select__ & match_kwargs("login", "password")
    default_groups = ("users",)

    def call(self, login, password, email=None, groups=None, **cwuserkwargs):
        cnx = self._cw
        if isinstance(password, str):
            # password should *always* be utf8 encoded
            password = password.encode("UTF8")
        cwuserkwargs["login"] = login
        cwuserkwargs["upassword"] = password
        # we have to create the user
        user = cnx.create_entity("CWUser", **cwuserkwargs)
        if groups is None:
            groups = self.default_groups
        assert groups, "CWUsers must belong to at least one CWGroup"
        group_names = ", ".join(f"{group!r}" for group in groups)
        cnx.execute(
            "SET X in_group G WHERE X eid %%(x)s, G name IN (%s)" % group_names,
            {"x": user.eid},
        )
        if email or "@" in login:
            d = {"login": login, "email": email or login}
            cnx.execute(
                "INSERT EmailAddress X: X address %(email)s, "
                "U primary_email X, U use_email X "
                "WHERE U login %(login)s",
                d,
                build_descr=False,
            )

        return user


class SourceSynchronizationService(Service):
    """Force synchronization of a datafeed source. Actual synchronization is done
    asynchronously, this will simply create and return the entity which will hold the import
    log.
    """

    __regid__ = "source-sync"
    __select__ = Service.__select__ & match_user_groups("managers")

    def call(self, source_eid):
        source = self._cw.repo.source_by_eid(source_eid)
        result = source.pull_data(self._cw, force=True, sync=False)
        return result["import_log_eid"]



from yams.buildobjs import EntityType, String, SubjectRelation, RelationType


class Tag(EntityType):
    """tags are used by users to mark entities.
    When you include the Tag entity, all application specific entities
    may then be tagged using the "tags" relation.
    """

    name = String(required=True, fulltextindexed=True, unique=True, maxsize=128)
    # when using this component, add the Tag tag X relation for each type that
    # should be taggeable
    tags = SubjectRelation("Tag", description="tagged objects")


class tags(RelationType):
    """indicates that an entity is classified by a given tag"""



#!/usr/bin/env python3

from logilab.common.configuration import REQUIRED

options = (
    (
        "a-required-string-option",
        {
            "type": "string",
            "default": REQUIRED,
            "help": "",
            "group": "test",
            "level": 1,
        },
    ),
    (
        "a-required-int-option",
        {
            "type": "int",
            "default": REQUIRED,
            "help": "",
            "group": "test",
        },
    ),
    (
        "a-required-yn-option",
        {
            "type": "yn",
            "default": REQUIRED,
            "help": "",
            "group": "test",
        },
    ),
)



# copyright 2004-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of cubicweb.
#
# yams is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# yams is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with yams. If not, see <https://www.gnu.org/licenses/>.
"""write a schema as sql"""

from hashlib import md5

from logilab import database
from logilab.common.decorators import monkeypatch
from yams.constraints import SizeConstraint, UniqueConstraint, Attribute, NOW, TODAY

from cubicweb.schema import constraint_name_for

# default are usually not handled at the sql level. If you want them, set
# SET_DEFAULT to True
SET_DEFAULT = False


# backport fix for lgdb #6662663
@monkeypatch(database._GenericAdvFuncHelper)
def sql_create_index(self, table, column, unique=False):
    idx = self._index_name(table, column, unique)
    if unique:
        return f"ALTER TABLE {table} ADD CONSTRAINT {idx} UNIQUE({column})"
    else:
        return f"CREATE INDEX {idx} ON {table}({column})"


@monkeypatch(database._GenericAdvFuncHelper)
def _index_name(self, table, column, unique=False):
    if unique:
        return build_index_name(table, [column], prefix="key_")
    else:
        return build_index_name(table, [column], prefix="idx_")


def build_index_name(table, columns, prefix="idx_"):
    """Return a predictable-but-size-constrained name for an index on `table(*columns)`, using an
    md5 hash.
    """
    return "{}{}".format(
        prefix,
        md5((table + "," + ",".join(sorted(columns))).encode("ascii")).hexdigest(),
    )


def rschema_has_table(rschema, skip_relations):
    """Return True if the given schema should have a table in the database."""
    return not (
        rschema.final
        or rschema.inlined
        or rschema.rule
        or rschema.type in skip_relations
    )


def schema2sql(dbhelper, schema, skip_entities=(), skip_relations=(), prefix=""):
    """Yield SQL statements to create a database schema for the given Yams schema.

    `prefix` may be a string that will be prepended to all table / column names (usually, 'cw_').
    """
    for etype in sorted(schema.entities()):
        eschema = schema.entity_schema_for(etype)
        if eschema.final or eschema.type in skip_entities:
            continue
        yield from eschema2sql(dbhelper, eschema, skip_relations, prefix)
    for rtype in sorted(schema.relations()):
        rschema = schema.relation_schema_for(rtype)
        if rschema_has_table(rschema, skip_relations):
            yield from rschema2sql(rschema)


def unique_index_name(eschema, attrs):
    """Return a predictable-but-size-constrained name for a multi-columns unique index on
    given attributes of the entity schema (actually, the later may be a schema or a string).
    """
    # keep giving eschema instead of table name for bw compat
    table = str(eschema)
    # unique_index_name is used as name of CWUniqueConstraint, hence it should be unicode
    return build_index_name(table, attrs, "unique_")


def iter_unique_index_names(eschema):
    """Yield (attrs, index name) where attrs is a list of entity type's attribute names that should
    be unique together, and index name the unique index name.
    """
    for attrs in eschema._unique_together or ():
        yield attrs, unique_index_name(eschema, attrs)


def eschema_sql_def(dbhelper, eschema, skip_relations=(), prefix=""):
    """Return a list of (column names, sql type def) for the given entity schema.

    No constraint nor index are considered - this function is usually for massive import purpose.
    """
    attrs = [
        attrdef
        for attrdef in eschema.attribute_definitions()
        if attrdef[0].type not in skip_relations
    ]
    attrs += [
        (rschema, None)
        for rschema in list(eschema.subject_relations.values())
        if not rschema.final and rschema.inlined
    ]
    result = []
    for i in range(len(attrs)):
        rschema, attrschema = attrs[i]
        if attrschema is not None:
            # creating = False will avoid NOT NULL / REFERENCES constraints
            sqltype = aschema2sql(
                dbhelper, eschema, rschema, attrschema, creating=False
            )
        else:  # inline relation
            sqltype = "integer"
        result.append((f"{prefix}{rschema.type}", sqltype))
    return result


def eschema2sql(dbhelper, eschema, skip_relations=(), prefix=""):
    """Yield SQL statements to initialize database from an entity schema."""
    table = prefix + eschema.type
    output = []
    w = output.append
    w(f"CREATE TABLE {table}(")
    attrs = [
        attrdef
        for attrdef in eschema.attribute_definitions()
        if attrdef[0].type not in skip_relations
    ]
    attrs += [
        (rschema, None)
        for rschema in list(eschema.subject_relations.values())
        if not rschema.final and rschema.inlined
    ]
    for i in range(len(attrs)):
        rschema, attrschema = attrs[i]
        if attrschema is not None:
            sqltype = aschema2sql(dbhelper, eschema, rschema, attrschema)
        else:  # inline relation
            sqltype = "integer REFERENCES entities (eid)"
        if i == len(attrs) - 1:
            w(f" {prefix}{rschema.type} {sqltype}")
        else:
            w(f" {prefix}{rschema.type} {sqltype},")
    for rschema, aschema in attrs:
        if aschema is None:  # inline relation
            continue
        rdef = rschema.relation_definition(eschema.type, aschema.type)
        for constraint in rdef.constraints:
            cstrname, check = check_constraint(
                rdef, constraint, dbhelper, prefix=prefix
            )
            if cstrname is not None:
                w(f", CONSTRAINT {cstrname} CHECK({check})")
    w(")")
    yield "\n".join(output)
    # create indexes
    for i in range(len(attrs)):
        rschema, attrschema = attrs[i]
        if attrschema is None or eschema.relation_definition(rschema).indexed:
            yield dbhelper.sql_create_index(table, prefix + rschema.type)
        if attrschema and any(
            isinstance(cstr, UniqueConstraint)
            for cstr in eschema.relation_definition(rschema).constraints
        ):
            yield dbhelper.sql_create_index(table, prefix + rschema.type, unique=True)
    for attrs, index_name in iter_unique_index_names(eschema):
        columns = [f"{prefix}{attr}" for attr in attrs]
        sqls = dbhelper.sqls_create_multicol_unique_index(table, columns, index_name)
        for sql in sqls:
            yield sql.rstrip(";")  # remove trailing ';' for consistency


def constraint_value_as_sql(value, dbhelper, prefix):
    """Return the SQL value from a Yams constraint's value, handling special cases where it's a
    `Attribute`, `TODAY` or `NOW` instance instead of a literal value.
    """
    if isinstance(value, Attribute):
        return prefix + value.attr
    elif isinstance(value, TODAY):
        return dbhelper.sql_current_date()
    elif isinstance(value, NOW):
        return dbhelper.sql_current_timestamp()
    else:
        # XXX more quoting for literals?
        return value


def check_constraint(rdef, constraint, dbhelper, prefix=""):
    """Return (constraint name, constraint SQL definition) for the given relation definition's
    constraint. Maybe (None, None) if the constraint is not handled in the backend.
    """
    attr = rdef.relation_type.type
    cstrname = constraint_name_for(constraint, rdef)
    if constraint.type() == "BoundaryConstraint":
        value = constraint_value_as_sql(constraint.boundary, dbhelper, prefix)
        return cstrname, f"{prefix}{attr} {constraint.operator} {value}"
    elif constraint.type() == "IntervalBoundConstraint":
        condition = []
        if constraint.minvalue is not None:
            value = constraint_value_as_sql(constraint.minvalue, dbhelper, prefix)
            condition.append(f"{prefix}{attr} >= {value}")
        if constraint.maxvalue is not None:
            value = constraint_value_as_sql(constraint.maxvalue, dbhelper, prefix)
            condition.append(f"{prefix}{attr} <= {value}")
        return cstrname, " AND ".join(condition)
    elif constraint.type() == "StaticVocabularyConstraint":
        sample = next(iter(constraint.vocabulary()))
        if not isinstance(sample, str):
            values = ", ".join(str(word) for word in constraint.vocabulary())
        else:
            # XXX better quoting?
            values = ", ".join(
                "'%s'" % word.replace("'", "''") for word in constraint.vocabulary()
            )
        return cstrname, f"{prefix}{attr} IN ({values})"
    return None, None


def aschema2sql(dbhelper, eschema, rschema, aschema, creating=True):
    """Return string containing a SQL table's column definition from attribute schema."""
    attr = rschema.type
    rdef = rschema.relation_definition(eschema.type, aschema.type)
    sqltype = type_from_rdef(dbhelper, rdef)
    if SET_DEFAULT:
        default = eschema.default(attr)
        if default is not None:
            if aschema.type == "Boolean":
                sqltype += f" DEFAULT {dbhelper.boolean_value(default)}"
            elif aschema.type == "String":
                sqltype += f" DEFAULT {str(default)!r}"
            elif aschema.type in ("Int", "BigInt", "Float"):
                sqltype += f" DEFAULT {default}"
            # XXX ignore default for other type
            # this is expected for NOW / TODAY
    if creating:
        if rdef.uid:
            sqltype += " PRIMARY KEY REFERENCES entities (eid)"
        elif rdef.cardinality[0] == "1":
            # don't set NOT NULL if backend isn't able to change it later
            if dbhelper.alter_column_support:
                sqltype += " NOT NULL"
    # else we're getting sql type to alter a column, we don't want key / indexes
    # / null modifiers
    return sqltype


def type_from_rdef(dbhelper, rdef):
    """Return a string containing SQL type name for the given relation definition."""
    constraints = list(rdef.constraints)
    sqltype = None
    if rdef.object.type == "String":
        for constraint in constraints:
            if isinstance(constraint, SizeConstraint) and constraint.max is not None:
                size_constrained_string = dbhelper.TYPE_MAPPING.get(
                    "SizeConstrainedString", "varchar(%s)"
                )
                sqltype = size_constrained_string % constraint.max
                break
    if sqltype is None:
        sqltype = sql_type(dbhelper, rdef)
    return sqltype


def sql_type(dbhelper, rdef):
    """
    Return a string containing SQL type to use to store values of the given relation definition.
    """
    sqltype = dbhelper.TYPE_MAPPING[rdef.object]
    if callable(sqltype):
        sqltype = sqltype(rdef)
    return sqltype


_SQL_SCHEMA = """
CREATE TABLE %(table)s (
  eid_from INTEGER NOT NULL REFERENCES entities (eid),
  eid_to INTEGER NOT NULL REFERENCES entities (eid),
  CONSTRAINT %(pkey_idx)s PRIMARY KEY(eid_from, eid_to)
);

CREATE INDEX %(from_idx)s ON %(table)s(eid_from);
CREATE INDEX %(to_idx)s ON %(table)s(eid_to)"""


def rschema2sql(rschema):
    """Yield SQL statements to create database table and indexes for a Yams relation schema."""
    assert not rschema.rule
    table = f"{rschema.type}_relation"
    sqls = _SQL_SCHEMA % {
        "table": table,
        "pkey_idx": build_index_name(table, ["eid_from", "eid_to"], "key_"),
        "from_idx": build_index_name(table, ["eid_from"], "idx_"),
        "to_idx": build_index_name(table, ["eid_to"], "idx_"),
    }
    for sql in sqls.split(";"):
        yield sql.strip()


def grant_schema(schema, user, set_owner=True, skip_entities=(), prefix=""):
    """Yield SQL statements to give all access (and ownership if `set_owner` is True) on the
    database tables for the given Yams schema to `user`.

    `prefix` may be a string that will be prepended to all table / column names (usually, 'cw_').
    """
    for etype in sorted(schema.entities()):
        eschema = schema.entity_schema_for(etype)
        if eschema.final or etype in skip_entities:
            continue
        yield from grant_eschema(eschema, user, set_owner, prefix=prefix)
    for rtype in sorted(schema.relations()):
        rschema = schema.relation_schema_for(rtype)
        if rschema_has_table(
            rschema, skip_relations=()
        ):  # XXX skip_relations should be specified
            yield from grant_rschema(rschema, user, set_owner)


def grant_eschema(eschema, user, set_owner=True, prefix=""):
    """Yield SQL statements to give all access (and ownership if `set_owner` is True) on the
    database tables for the given Yams entity schema to `user`.
    """
    etype = eschema.type
    if set_owner:
        yield f"ALTER TABLE {prefix}{etype} OWNER TO {user}"
    yield f"GRANT ALL ON {prefix}{etype} TO {user}"


def grant_rschema(rschema, user, set_owner=True):
    """Yield SQL statements to give all access (and ownership if `set_owner` is True) on the
    database tables for the given Yams relation schema to `user`.
    """
    if set_owner:
        yield f"ALTER TABLE {rschema.type}_relation OWNER TO {user}"
    yield f"GRANT ALL ON {rschema.type}_relation TO {user}"



# copyright 2012 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""Tests for cubicweb.server.utils module."""

import sched

from cubicweb.devtools import testlib
from cubicweb.server import utils


class UtilsTC(testlib.BaseTestCase):
    def test_crypt(self):
        for hash in (
            utils.crypt_password("xxx"),  # default sha512
            b"ab$5UsKFxRKKN.d8iBIFBnQ80",  # custom md5
            b"ab4Vlm81ZUHlg",  # DES
        ):
            self.assertEqual(utils.crypt_password("xxx", hash), hash)
            self.assertEqual(utils.crypt_password("xxx", hash), hash)
            self.assertEqual(
                utils.crypt_password("xxx", hash.decode("ascii")), hash.decode("ascii")
            )
            self.assertEqual(utils.crypt_password("yyy", hash), b"")

        # accept any password for empty hashes (is it a good idea?)
        self.assertEqual(utils.crypt_password("xxx", ""), "")
        self.assertEqual(utils.crypt_password("yyy", ""), "")

    def test_schedule_periodic_task(self):
        scheduler = sched.scheduler()
        this = []

        def fill_this(x):
            this.append(x)
            if len(this) > 2:
                raise SystemExit()
            elif len(this) > 1:
                raise RuntimeError()

        event = utils.schedule_periodic_task(scheduler, 0.01, fill_this, 1)
        self.assertEqual(event.action.__name__, "fill_this")
        self.assertEqual(len(scheduler.queue), 1)

        with self.assertLogs("cubicweb.scheduler", level="ERROR") as cm:
            scheduler.run()
        self.assertEqual(this, [1] * 3)
        self.assertEqual(len(cm.output), 2)
        self.assertIn('Unhandled exception in periodic task "fill_this"', cm.output[0])
        self.assertIn('"fill_this" not re-scheduled', cm.output[1])


if __name__ == "__main__":
    import unittest

    unittest.main()



# copyright 2003-2015 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.

from datetime import datetime, timezone
from threading import Thread

import logilab.database as lgdb
from cubicweb.server.test.unittest_querier import FixedOffset

from yams import ValidationError
from cubicweb.devtools import (
    startpgcluster,
    stoppgcluster,
)
from cubicweb.devtools.testlib import CubicWebTC
from cubicweb.devtools.apptest_config import PostgresApptestConfiguration
from cubicweb.entities.adapters import IFTIndexableAdapter
from cubicweb.predicates import is_instance


def setUpModule():
    startpgcluster(__file__)


def tearDownModule():
    stoppgcluster(__file__)


class PostgresTimeoutConfiguration(PostgresApptestConfiguration):
    def __init__(self, *args, **kwargs):
        self.default_sources = PostgresApptestConfiguration.default_sources.copy()
        self.default_sources["system"] = PostgresApptestConfiguration.default_sources[
            "system"
        ].copy()
        self.default_sources["system"]["db-statement-timeout"] = 200
        super().__init__(*args, **kwargs)


class PostgresFTITC(CubicWebTC):
    configcls = PostgresApptestConfiguration

    def test_eid_range(self):
        # concurrent allocation of eid ranges
        source = self.repo.system_source
        range1 = []
        range2 = []

        def allocate_eid_ranges(session, target):
            for x in range(1, 10):
                eid = source.create_eid(session, count=x)
                target.extend(range(eid - x, eid))

        t1 = Thread(target=lambda: allocate_eid_ranges(self.session, range1))
        t2 = Thread(target=lambda: allocate_eid_ranges(self.session, range2))
        t1.start()
        t2.start()
        t1.join()
        t2.join()
        self.assertEqual(range1, sorted(range1))
        self.assertEqual(range2, sorted(range2))
        self.assertEqual(set(), set(range1) & set(range2))

    def test_occurence_count(self):
        with self.admin_access.repo_cnx() as cnx:
            c1 = cnx.create_entity(
                "Card", title="c1", content="cubicweb cubicweb cubicweb"
            )
            c2 = cnx.create_entity("Card", title="c3", content="cubicweb")
            c3 = cnx.create_entity("Card", title="c2", content="cubicweb cubicweb")
            cnx.commit()
            self.assertEqual(
                cnx.execute(
                    "Card X ORDERBY FTIRANK(X) DESC " 'WHERE X has_text "cubicweb"'
                ).rows,
                [
                    [
                        c1.eid,
                    ],
                    [
                        c3.eid,
                    ],
                    [
                        c2.eid,
                    ],
                ],
            )

    def test_attr_weight(self):
        class CardIFTIndexableAdapter(IFTIndexableAdapter):
            __select__ = is_instance("Card")
            attr_weight = {"title": "A"}

        with self.temporary_appobjects(CardIFTIndexableAdapter):
            with self.admin_access.repo_cnx() as cnx:
                c1 = cnx.create_entity(
                    "Card", title="c1", content="cubicweb cubicweb cubicweb"
                )
                c2 = cnx.create_entity("Card", title="c2", content="cubicweb cubicweb")
                c3 = cnx.create_entity("Card", title="cubicweb", content="autre chose")
                cnx.commit()
                self.assertEqual(
                    cnx.execute(
                        "Card X ORDERBY FTIRANK(X) DESC " 'WHERE X has_text "cubicweb"'
                    ).rows,
                    [
                        [
                            c3.eid,
                        ],
                        [
                            c1.eid,
                        ],
                        [
                            c2.eid,
                        ],
                    ],
                )

    def test_entity_weight(self):
        class PersonneIFTIndexableAdapter(IFTIndexableAdapter):
            __select__ = is_instance("Personne")
            entity_weight = 2.0

        with self.temporary_appobjects(PersonneIFTIndexableAdapter):
            with self.admin_access.repo_cnx() as cnx:
                c1 = cnx.create_entity("Personne", nom="c1", prenom="cubicweb")
                c2 = cnx.create_entity(
                    "Comment", content="cubicweb cubicweb", comments=c1
                )
                c3 = cnx.create_entity(
                    "Comment", content="cubicweb cubicweb cubicweb", comments=c1
                )
                cnx.commit()
                self.assertEqual(
                    cnx.execute(
                        "Any X ORDERBY FTIRANK(X) DESC " 'WHERE X has_text "cubicweb"'
                    ).rows,
                    [
                        [
                            c1.eid,
                        ],
                        [
                            c3.eid,
                        ],
                        [
                            c2.eid,
                        ],
                    ],
                )

    def test_tz_datetime(self):
        with self.admin_access.repo_cnx() as cnx:
            bob = cnx.create_entity(
                "Personne",
                nom="bob",
                tzdatenaiss=datetime(1977, 6, 7, 2, 0, tzinfo=FixedOffset(1)),
            )
            datenaiss = cnx.execute("Any XD WHERE X nom 'bob', X tzdatenaiss XD")[0][0]
            self.assertIsNotNone(datenaiss.tzinfo)
            self.assertEqual(datenaiss.utctimetuple()[:5], (1977, 6, 7, 1, 0))
            cnx.commit()
            cnx.create_entity(
                "Personne",
                nom="boby",
                tzdatenaiss=datetime(1977, 6, 7, 2, 0, tzinfo=timezone.utc),
            )
            datenaiss = cnx.execute("Any XD WHERE X nom 'boby', X tzdatenaiss XD")[0][0]
            self.assertIsNotNone(datenaiss.tzinfo)
            self.assertEqual(datenaiss.utctimetuple()[:5], (1977, 6, 7, 2, 0))
            rset = cnx.execute(
                "Any X WHERE X tzdatenaiss %(d)s",
                {"d": datetime(1977, 6, 7, 2, 0, tzinfo=FixedOffset(1))},
            )
            self.assertEqual(rset.rows, [[bob.eid]])

    def test_constraint_validationerror(self):
        with self.admin_access.repo_cnx() as cnx:
            with cnx.allow_all_hooks_but("integrity"):
                with self.assertRaises(ValidationError) as cm:
                    cnx.execute("INSERT Note N: N type 'nogood'")
                self.assertEqual(
                    cm.exception.errors,
                    {
                        "type-subject": "invalid value %(KEY-value)s, it must be one of %(KEY-choices)s"
                    },
                )
                self.assertEqual(
                    cm.exception.msgargs,
                    {
                        "type-subject-value": '"nogood"',
                        "type-subject-choices": '"todo", "a", "b", "T", "lalala"',
                    },
                )


class PostgresStatementTimeoutTC(CubicWebTC):
    configcls = PostgresTimeoutConfiguration

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.orig_connect_hooks = lgdb.SQL_CONNECT_HOOKS["postgres"][:]

    @classmethod
    def tearDownClass(cls):
        lgdb.SQL_CONNECT_HOOKS["postgres"] = cls.orig_connect_hooks

    def test_statement_timeout(self):
        with self.admin_access.repo_cnx() as cnx:
            cnx.system_sql("select pg_sleep(0.1)")
            with self.assertRaises(Exception) as cm:
                cnx.system_sql("select pg_sleep(0.3)")
        self.assertIn("statement timeout", str(cm.exception))


class PostgresLimitSizeTC(CubicWebTC):
    configcls = PostgresApptestConfiguration

    def test(self):
        with self.admin_access.repo_cnx() as cnx:

            def sql(string):
                return cnx.system_sql(string).fetchone()[0]

            for html, size, expected in [
                ("<p>hello</p>", 20, "<p>hello</p>"),
                ("<p>hello</p>", 2, "he..."),
                ("<br/>hello", 2, "he..."),
                ('<span class="1">he</span>llo', 2, "he..."),
                ("<span>a>b</span>", 2, "a>..."),
            ]:
                with self.subTest(html=html, size=size):
                    actual = sql(
                        "SELECT limit_size('%s', 'text/html', %d)" % (html, size)
                    )
                    self.assertEqual(actual, expected)


class PostgresGroupConcatTC(CubicWebTC):
    configcls = PostgresApptestConfiguration

    def test(self):
        with self.admin_access.repo_cnx() as cnx:
            cnx.create_entity("Personne", nom="ABC_1", prenom="Alice")
            cnx.create_entity("Personne", nom="ABC_2", prenom="Alice")
            cnx.create_entity("Personne", nom="ABC", prenom="Bob")
            cnx.create_entity("Personne", nom="ABC")

            cnx.commit()

            firstnames_rset = cnx.execute(
                "Any GROUP_CONCAT(P) WHERE X is Personne, X prenom P"
            )

            # spy's firstname is null, it should not appear in the result
            self.assertEqual(list(firstnames_rset), [["Alice, Bob"]])

            lastnames_rset = cnx.execute(
                "Any GROUP_CONCAT(N) WHERE X is Personne, X nom N"
            )

            # ABC should appear only once
            self.assertEqual(list(lastnames_rset), [["ABC_1, ABC_2, ABC"]])


if __name__ == "__main__":
    import unittest

    unittest.main()



"""This module provide syntaxe highlight functions"""

from logilab.common.logging_ext import _colorable_terminal

try:
    from pygments import highlight as pygments_highlight
    from pygments.lexers import get_lexer_by_name
    from pygments.formatters.terminal import TerminalFormatter
    from pygments.formatters.html import HtmlFormatter

    has_pygments = True
except ImportError:
    has_pygments = False


def highlight_terminal(code, language):
    if not has_pygments:
        return code

    if not _colorable_terminal():
        return code

    return pygments_highlight(code, get_lexer_by_name(language), TerminalFormatter())


def highlight_html(code, language, linenos=False, linenostart=1, **kwargs):
    if not has_pygments:
        return str(code)

    return pygments_highlight(
        str(code),
        get_lexer_by_name(language),
        HtmlFormatter(
            wrapcode=True, linenos=linenos, linenostart=linenostart, **kwargs
        ),
    )


def generate_css():
    if has_pygments:
        return HtmlFormatter().get_style_defs()
    else:
        return ""



# copyright 2003-2014 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""RQL rewriting utilities : insert rql expression snippets into rql syntax
tree.

This is used for instance for read security checking in the repository.
"""

from logilab.common import tempattr
from logilab.common.graph import has_path
from rql import nodes as n, stmts, TypeResolverException
from rql.utils import common_parent
from yams import BadSchemaDefinition

from cubicweb import Unauthorized


def cleanup_solutions(rqlst, solutions):
    for sol in solutions:
        for vname in list(sol):
            if not (vname in rqlst.defined_vars or vname in rqlst.aliases):
                del sol[vname]


def add_types_restriction(schema, rqlst, newroot=None, solutions=None):
    if newroot is None:
        assert solutions is None
        if hasattr(rqlst, "_types_restr_added"):
            return
        solutions = rqlst.solutions
        newroot = rqlst
        rqlst._types_restr_added = True
    else:
        assert solutions is not None
        rqlst = rqlst.stmt
    allpossibletypes = {}
    for solution in solutions:
        for varname, etype in solution.items():
            # XXX not considering aliases by design, right ?
            if (
                varname not in newroot.defined_vars
                or schema.entity_schema_for(etype).final
            ):
                continue
            allpossibletypes.setdefault(varname, set()).add(etype)
    # XXX could be factorized with add_etypes_restriction from rql 0.31
    for varname in sorted(allpossibletypes):
        var = newroot.defined_vars[varname]
        stinfo = var.stinfo
        if stinfo.get("uidrel") is not None:
            continue  # eid specified, no need for additional type specification
        try:
            typerel = rqlst.defined_vars[varname].stinfo.get("typerel")
        except KeyError:
            assert varname in rqlst.aliases
            continue
        if newroot is rqlst and typerel is not None:
            mytyperel = typerel
        else:
            for vref in var.references():
                rel = vref.relation()
                if rel and rel.is_types_restriction():
                    mytyperel = rel
                    break
            else:
                mytyperel = None
        possibletypes = allpossibletypes[varname]
        if mytyperel is not None:
            if mytyperel.r_type == "is_instance_of":
                # turn is_instance_of relation into a is relation since we've
                # all possible solutions and don't want to bother with
                # potential is_instance_of incompatibility
                mytyperel.r_type = "is"
                if len(possibletypes) > 1:
                    node = n.Function("IN")
                    for etype in sorted(possibletypes):
                        node.append(n.Constant(etype, "etype"))
                else:
                    etype = next(iter(possibletypes))
                    node = n.Constant(etype, "etype")
                comp = mytyperel.children[1]
                comp.replace(comp.children[0], node)
            else:
                # variable has already some strict types restriction. new
                # possible types can only be a subset of existing ones, so only
                # remove no more possible types
                for cst in mytyperel.get_nodes(n.Constant):
                    if cst.value not in possibletypes:
                        cst.parent.remove(cst)
        else:
            # we have to add types restriction
            if stinfo.get("scope") is not None:
                rel = var.scope.add_type_restriction(var, possibletypes)
            else:
                # tree is not annotated yet, no scope set so add the restriction
                # to the root
                rel = newroot.add_type_restriction(var, possibletypes)
            stinfo["typerel"] = rel
        stinfo["possibletypes"] = possibletypes


def remove_solutions(origsolutions, solutions, defined):
    """when a rqlst has been generated from another by introducing security
    assertions, this method returns solutions which are contained in orig
    solutions
    """
    newsolutions = []
    for origsol in origsolutions:
        for newsol in solutions[:]:
            for var, etype in origsol.items():
                try:
                    if newsol[var] != etype:
                        try:
                            defined[var].stinfo["possibletypes"].remove(newsol[var])
                        except KeyError:
                            pass
                        break
                except KeyError:
                    # variable has been rewritten
                    continue
            else:
                newsolutions.append(newsol)
                solutions.remove(newsol)
    return newsolutions


def _add_noinvariant(noinvariant, restricted, select, nbtrees):
    # a variable can actually be invariant if it has not been restricted for
    # security reason or if security assertion hasn't modified the possible
    # solutions for the query
    for vname in restricted:
        try:
            var = select.defined_vars[vname]
        except KeyError:
            # this is an alias
            continue
        if nbtrees != 1 or len(var.stinfo["possibletypes"]) != 1:
            noinvariant.add(var)


def _expand_selection(terms, selected, aliases, select, newselect):
    for term in terms:
        for vref in term.iget_nodes(n.VariableRef):
            if vref.name not in selected:
                select.append_selected(vref)
                colalias = newselect.get_variable(vref.name, len(aliases))
                aliases.append(n.VariableRef(colalias))
                selected.add(vref.name)


def _has_multiple_cardinality(etypes, relation_definition, ttypes_func, cardindex):
    """return True if relation definitions from entity types (`etypes`) to
    target types returned by the `ttypes_func` function all have single (1 or ?)
    cardinality.
    """
    for etype in etypes:
        for ttype in ttypes_func(etype):
            if relation_definition(etype, ttype).cardinality[cardindex] in "+*":
                return True
    return False


def _compatible_relation(relations, stmt, sniprel):
    """Search among given rql relation nodes if there is one 'compatible' with the
    snippet relation, and return it if any, else None.

    A relation is compatible if it:
    * belongs to the currently processed statement,
    * isn't negged (i.e. direct parent is a NOT node)
    * isn't optional (outer join) or similarly as the snippet relation
    """
    for rel in relations:
        # don't share if relation's scope is not the current statement
        if rel.scope is not stmt:
            continue
        # don't share neged relation
        if rel.neged(strict=True):
            continue
        # don't share optional relation, unless the snippet relation is
        # similarly optional
        if rel.optional and rel.optional != sniprel.optional:
            continue
        return rel
    return None


def iter_relations(stinfo):
    # this is a function so that test may return relation in a predictable order
    return stinfo["relations"] - stinfo["rhsrelations"]


def need_exists(node):
    """Return true if the given node should be wrapped in an `Exists` node.

    This is true when node isn't already an `Exists` or `Not` node, nor a
    `And`/`Or` of `Exists` or `Not` nodes.
    """
    if isinstance(node, (n.Exists, n.Not)):
        return False
    if isinstance(node, (n.Or, n.And)):
        return need_exists(node.children[0]) or need_exists(node.children[1])
    return True


class Unsupported(Exception):
    """raised when an rql expression can't be inserted in some rql query
    because it create an unresolvable query (eg no solutions found)
    """


class VariableFromSubQuery(Exception):
    """flow control exception to indicate that a variable is coming from a
    subquery, and let parent act accordingly
    """

    def __init__(self, variable):
        self.variable = variable


class RQLRewriter:
    """Insert some rql snippets into another rql syntax tree, for security /
    relation vocabulary. This implies that it should only restrict results of
    the original query, not generate new ones. Hence, inserted snippets are
    inserted under an EXISTS node.

    This class *isn't thread safe*.
    """

    def __init__(self, session):
        self.session = session
        vreg = session.vreg
        self.schema = vreg.schema
        self.annotate = vreg.rqlhelper.annotate
        self._compute_solutions = vreg.compute_var_types

    def compute_solutions(self):
        self.annotate(self.select)
        try:
            self._compute_solutions(self.session, self.select, self.kwargs)
        except TypeResolverException:
            raise Unsupported(str(self.select))
        if len(self.select.solutions) < len(self.solutions):
            raise Unsupported()

    def insert_local_checks(self, select, kwargs, localchecks, restricted, noinvariant):
        """
        select: the rql syntax tree Select node
        kwargs: query arguments

        localchecks: {(('Var name', (rqlexpr1, rqlexpr2)),
                       ('Var name1', (rqlexpr1, rqlexpr23))): [solution]}

              (see querier._check_permissions docstring for more information)

        restricted: set of variable names to which an rql expression has to be
              applied

        noinvariant: set of variable names that can't be considered has
              invariant due to security reason (will be filed by this method)
        """
        nbtrees = len(localchecks)
        myunion = union = select.parent
        # transform in subquery when len(localchecks)>1 and groups
        if nbtrees > 1 and (
            select.orderby
            or select.groupby
            or select.having
            or select.has_aggregat
            or select.distinct
            or select.limit
            or select.offset
        ):
            newselect = stmts.Select()
            # only select variables in subqueries
            origselection = select.selection
            select.select_only_variables()
            select.has_aggregat = False
            # create subquery first so correct node are used on copy
            # (eg ColumnAlias instead of Variable)
            aliases = [
                n.VariableRef(newselect.get_variable(vref.name, i))
                for i, vref in enumerate(select.selection)
            ]
            selected = {vref.name for vref in aliases}
            # now copy original selection and groups
            for term in origselection:
                newselect.append_selected(term.copy(newselect))
            if select.orderby:
                sortterms = []
                for sortterm in select.orderby:
                    sortterms.append(sortterm.copy(newselect))
                    for fnode in sortterm.get_nodes(n.Function):
                        if fnode.name == "FTIRANK":
                            # we've to fetch the has_text relation as well
                            var = fnode.children[0].variable
                            rel = next(iter(var.stinfo["ftirels"]))
                            assert not rel.ored(), "unsupported"
                            newselect.add_restriction(rel.copy(newselect))
                            # remove relation from the orig select and
                            # cleanup variable stinfo
                            rel.parent.remove(rel)
                            var.stinfo["ftirels"].remove(rel)
                            var.stinfo["relations"].remove(rel)
                            # XXX not properly re-annotated after security insertion?
                            newvar = newselect.get_variable(var.name)
                            newvar.stinfo.setdefault("ftirels", set()).add(rel)
                            newvar.stinfo.setdefault("relations", set()).add(rel)
                newselect.set_orderby(sortterms)
                _expand_selection(select.orderby, selected, aliases, select, newselect)
                select.orderby = ()  # XXX dereference?
            if select.groupby:
                newselect.set_groupby([g.copy(newselect) for g in select.groupby])
                _expand_selection(select.groupby, selected, aliases, select, newselect)
                select.groupby = ()  # XXX dereference?
            if select.having:
                newselect.set_having([g.copy(newselect) for g in select.having])
                _expand_selection(select.having, selected, aliases, select, newselect)
                select.having = ()  # XXX dereference?
            if select.limit:
                newselect.limit = select.limit
                select.limit = None
            if select.offset:
                newselect.offset = select.offset
                select.offset = 0
            myunion = stmts.Union()
            newselect.set_with([n.SubQuery(aliases, myunion)], check=False)
            newselect.distinct = select.distinct
            solutions = [sol.copy() for sol in select.solutions]
            cleanup_solutions(newselect, solutions)
            newselect.set_possible_types(solutions)
            # if some solutions doesn't need rewriting, insert original
            # select as first union subquery
            if () in localchecks:
                myunion.append(select)
            # we're done, replace original select by the new select with
            # subqueries (more added in the loop below)
            union.replace(select, newselect)
        elif () not in localchecks:
            union.remove(select)
        for lcheckdef, lchecksolutions in localchecks.items():
            if not lcheckdef:
                continue
            myrqlst = select.copy(solutions=lchecksolutions)
            myunion.append(myrqlst)
            # in-place rewrite + annotation / simplification
            lcheckdef = [({v: "X"}, rqlexprs) for v, rqlexprs in lcheckdef]
            self.rewrite(myrqlst, lcheckdef, kwargs)
            _add_noinvariant(noinvariant, restricted, myrqlst, nbtrees)
        if () in localchecks:
            select.set_possible_types(localchecks[()])
            add_types_restriction(self.schema, select)
            _add_noinvariant(noinvariant, restricted, select, nbtrees)
        self.annotate(union)

    def rewrite(self, select, snippets, kwargs, existingvars=None):
        """
        snippets: (varmap, list of rql expression)
                  with varmap a *dict* {select var: snippet var}
        """
        self.select = select
        # remove_solutions used below require a copy
        self.solutions = solutions = select.solutions[:]
        self.kwargs = kwargs
        self.u_varname = None
        self.removing_ambiguity = False
        self.exists_snippet = {}
        self.pending_keys = []
        self.existingvars = existingvars
        # we have to annotate the rqlst before inserting snippets, even though
        # we'll have to redo it later
        self.annotate(select)
        self.insert_snippets(snippets)
        if not self.exists_snippet and self.u_varname:
            # U has been inserted than cancelled, cleanup
            select.undefine_variable(select.defined_vars[self.u_varname])
        # clean solutions according to initial solutions
        newsolutions = remove_solutions(
            solutions, select.solutions, select.defined_vars
        )
        assert len(newsolutions) >= len(solutions), (
            "rewritten rql %s has lost some solutions, there is probably "
            "something wrong in your schema permission (for instance using a "
            "RQLExpression which inserts a relation which doesn't exist in "
            "the schema)\nOrig solutions: %s\nnew solutions: %s"
            % (select, solutions, newsolutions)
        )
        if len(newsolutions) > len(solutions):
            newsolutions = self.remove_ambiguities(snippets, newsolutions)
            assert newsolutions
        select.solutions = newsolutions
        add_types_restriction(self.schema, select)

    def insert_snippets(self, snippets, varexistsmap=None):
        self.rewritten = {}
        for varmap, rqlexprs in snippets:
            if isinstance(varmap, dict):
                varmap = tuple(sorted(varmap.items()))
            else:
                assert isinstance(varmap, tuple), varmap
            if varexistsmap is not None and varmap not in varexistsmap:
                continue
            self.insert_varmap_snippets(varmap, rqlexprs, varexistsmap)

    def init_from_varmap(self, varmap, varexistsmap=None):
        self.varmap = varmap
        self.revvarmap = {}
        self.varinfos = []
        for i, (selectvar, snippetvar) in enumerate(varmap):
            assert snippetvar in "SOX"
            self.revvarmap[snippetvar] = (selectvar, i)
            vi = {}
            self.varinfos.append(vi)
            try:
                vi["const"] = int(selectvar)
                vi["rhs_rels"] = vi["lhs_rels"] = {}
            except ValueError:
                try:
                    vi["stinfo"] = sti = self.select.defined_vars[selectvar].stinfo
                except KeyError:
                    vi["stinfo"] = sti = self._subquery_variable(selectvar)
                if varexistsmap is None:
                    # build an index for quick access to relations
                    vi["rhs_rels"] = {}
                    for rel in sti.get("rhsrelations", []):
                        vi["rhs_rels"].setdefault(rel.r_type, []).append(rel)
                    vi["lhs_rels"] = {}
                    for rel in sti.get("relations", []):
                        if rel not in sti.get("rhsrelations", []):
                            vi["lhs_rels"].setdefault(rel.r_type, []).append(rel)
                else:
                    vi["rhs_rels"] = vi["lhs_rels"] = {}

    def _subquery_variable(self, selectvar):
        raise VariableFromSubQuery(selectvar)

    def insert_varmap_snippets(self, varmap, rqlexprs, varexistsmap):
        try:
            self.init_from_varmap(varmap, varexistsmap)
        except VariableFromSubQuery as ex:
            # variable may have been moved to a newly inserted subquery
            # we should insert snippet in that subquery
            subquery = self.select.aliases[ex.variable].query
            assert len(subquery.children) == 1, subquery
            subselect = subquery.children[0]
            RQLRewriter(self.session).rewrite(
                subselect, [(varmap, rqlexprs)], self.kwargs
            )
            return
        self._insert_scope = None
        previous = None
        inserted = False
        for rqlexpr in rqlexprs:
            self.current_expr = rqlexpr
            if varexistsmap is None:
                try:
                    new = self.insert_snippet(varmap, rqlexpr.snippet_rqlst, previous)
                except Unsupported:
                    continue
                inserted = True
                if new is not None and self._insert_scope is None:
                    self.exists_snippet[rqlexpr] = new
                previous = previous or new
            else:
                # called to reintroduce snippet due to ambiguity creation,
                # so skip snippets which are not introducing this ambiguity
                exists = varexistsmap[varmap]
                if self.exists_snippet.get(rqlexpr) is exists:
                    self.insert_snippet(varmap, rqlexpr.snippet_rqlst, exists)
        if varexistsmap is None and not inserted:
            # no rql expression found matching rql solutions. User has no access right
            raise Unauthorized()  # XXX may also be because of bad constraints in schema definition

    def insert_snippet(self, varmap, snippetrqlst, previous=None):
        new = snippetrqlst.where.accept(self)
        existing = self.existingvars
        self.existingvars = None
        try:
            return self._insert_snippet(varmap, previous, new)
        finally:
            self.existingvars = existing

    def _inserted_root(self, new):
        if need_exists(new):
            new = n.Exists(new)
        return new

    def _insert_snippet(self, varmap, previous, new):
        """insert `new` snippet into the syntax tree, which have been rewritten
        using `varmap`. In cases where an action is protected by several rql
        expresssion, `previous` will be the first rql expression which has been
        inserted, and so should be ORed with the following expressions.
        """
        if new is not None:
            if self._insert_scope is None:
                insert_scope = None
                for vi in self.varinfos:
                    scope = vi.get("stinfo", {}).get("scope", self.select)
                    if insert_scope is None:
                        insert_scope = scope
                    else:
                        insert_scope = common_parent(scope, insert_scope)
            else:
                insert_scope = self._insert_scope
            if self._insert_scope is None and any(
                vi.get("stinfo", {}).get("optrelations") for vi in self.varinfos
            ):
                assert previous is None
                self._insert_scope, new = self.snippet_subquery(varmap, new)
                self.insert_pending()
                return new
            new = self._inserted_root(new)
            if previous is None:
                insert_scope.add_restriction(new)
            else:
                grandpa = previous.parent
                or_ = n.Or(previous, new)
                grandpa.replace(previous, or_)
            if not self.removing_ambiguity:
                try:
                    self.compute_solutions()
                except Unsupported:
                    # some solutions have been lost, can't apply this rql expr
                    if previous is None:
                        self.current_statement().remove_node(new, undefine=True)
                    else:
                        grandpa.replace(or_, previous)
                        self._cleanup_inserted(new)
                    raise
                else:
                    with tempattr(self, "_insert_scope", new):
                        self.insert_pending()
            return new
        self.insert_pending()

    def insert_pending(self):
        """pending_keys hold variable referenced by U has_<action>_permission X
        relation.

        Once the snippet introducing this has been inserted and solutions
        recomputed, we have to insert snippet defined for <action> of entity
        types taken by X
        """
        stmt = self.current_statement()
        while self.pending_keys:
            key, action = self.pending_keys.pop()
            try:
                varname = self.rewritten[key]
            except KeyError:
                try:
                    varname = self.revvarmap[key[-1]][0]
                except KeyError:
                    # variable isn't used anywhere else, we can't insert security
                    raise Unauthorized()
            ptypes = stmt.defined_vars[varname].stinfo["possibletypes"]
            if len(ptypes) > 1:
                # XXX dunno how to handle this
                self.session.error(
                    "cant check security of %s, ambigous type for %s in %s",
                    stmt,
                    varname,
                    key[0],
                )  # key[0] == the rql expression
                raise Unauthorized()
            etype = next(iter(ptypes))
            eschema = self.schema.entity_schema_for(etype)
            if not eschema.has_perm(self.session, action):
                rqlexprs = eschema.get_rqlexprs(action)
                if not rqlexprs:
                    raise Unauthorized()
                self.insert_snippets([({varname: "X"}, rqlexprs)])

    def snippet_subquery(self, varmap, transformedsnippet):
        """introduce the given snippet in a subquery"""
        subselect = stmts.Select()
        snippetrqlst = n.Exists(transformedsnippet.copy(subselect))
        get_rschema = self.schema.relation_schema_for
        aliases = []
        done = set()
        for i, (selectvar, _) in enumerate(varmap):
            need_null_test = False
            subselectvar = subselect.get_variable(selectvar)
            subselect.append_selected(n.VariableRef(subselectvar))
            aliases.append(selectvar)
            todo = [(selectvar, self.varinfos[i]["stinfo"])]
            while todo:
                varname, stinfo = todo.pop()
                done.add(varname)
                for rel in iter_relations(stinfo):
                    if rel in done:
                        continue
                    done.add(rel)
                    rschema = get_rschema(rel.r_type)
                    if rschema.final or rschema.inlined:
                        subselect_vrefs = []
                        rel.children[0].name = varname  # XXX explain why
                        subselect.add_restriction(rel.copy(subselect))
                        for vref in rel.children[1].iget_nodes(n.VariableRef):
                            if isinstance(vref.variable, n.ColumnAlias):
                                # XXX could probably be handled by generating the
                                # subquery into the detected subquery
                                raise BadSchemaDefinition(
                                    "cant insert security because of usage two inlined "
                                    "relations in this query. You should probably at "
                                    "least uninline %s" % rel.r_type
                                )
                            subselect.append_selected(vref.copy(subselect))
                            aliases.append(vref.name)
                            subselect_vrefs.append(vref)
                        self.select.remove_node(rel)
                        # when some inlined relation has to be copied in the
                        # subquery and that relation is optional, we need to
                        # test that either value is NULL or that the snippet
                        # condition is satisfied
                        if varname == selectvar and rel.optional and rschema.inlined:
                            need_null_test = True
                        # also, if some attributes or inlined relation of the
                        # object variable are accessed, we need to get all those
                        # from the subquery as well
                        for vref in subselect_vrefs:
                            if vref.name not in done and rschema.inlined:
                                # we can use vref here define in above for loop
                                ostinfo = vref.variable.stinfo
                                for orel in iter_relations(ostinfo):
                                    orschema = get_rschema(orel.r_type)
                                    if orschema.final or orschema.inlined:
                                        todo.append((vref.name, ostinfo))
                                        break
            if need_null_test:
                snippetrqlst = n.Or(
                    n.make_relation(
                        subselect.get_variable(selectvar),
                        "is",
                        (None, None),
                        n.Constant,
                        operator="=",
                    ),
                    snippetrqlst,
                )
        subselect.add_restriction(snippetrqlst)
        if self.u_varname:
            # generate an identifier for the substitution
            argname = subselect.allocate_varname()
            while argname in self.kwargs:
                argname = subselect.allocate_varname()
            subselect.add_constant_restriction(
                subselect.get_variable(self.u_varname), "eid", argname, "Substitute"
            )
            self.kwargs[argname] = self.session.user.eid
        add_types_restriction(
            self.schema, subselect, subselect, solutions=self.solutions
        )
        myunion = stmts.Union()
        myunion.append(subselect)
        aliases = [
            n.VariableRef(self.select.get_variable(name, i))
            for i, name in enumerate(aliases)
        ]
        self.select.add_subquery(n.SubQuery(aliases, myunion), check=False)
        self._cleanup_inserted(transformedsnippet)
        try:
            self.compute_solutions()
        except Unsupported:
            # some solutions have been lost, can't apply this rql expr
            self.select.remove_subquery(self.select.with_[-1])
            raise
        return subselect, snippetrqlst

    def remove_ambiguities(self, snippets, newsolutions):
        # the snippet has introduced some ambiguities, we have to resolve them
        # "manually"
        variantes = self.build_variantes(newsolutions)
        # if all ambiguities have been generated by variables within a "NOT
        # EXISTS()#" or with type explicitly specified, we've nothing to change
        if not variantes:
            return newsolutions
        # insert "is" where necessary
        varexistsmap = {}
        self.removing_ambiguity = True
        for (erqlexpr, varmap, oldvarname), etype in variantes[0].items():
            varname = self.rewritten[(erqlexpr, varmap, oldvarname)]
            var = self.select.defined_vars[varname]
            exists = var.references()[0].scope
            exists.add_constant_restriction(var, "is", etype, "etype")
            varexistsmap[varmap] = exists
        # insert ORED exists where necessary
        for variante in variantes[1:]:
            self.insert_snippets(snippets, varexistsmap)
            for key, etype in variante.items():
                varname = self.rewritten[key]
                try:
                    var = self.select.defined_vars[varname]
                except KeyError:
                    # not a newly inserted variable
                    continue
                exists = var.references()[0].scope
                exists.add_constant_restriction(var, "is", etype, "etype")
        # recompute solutions
        self.compute_solutions()
        # clean solutions according to initial solutions
        return remove_solutions(
            self.solutions, self.select.solutions, self.select.defined_vars
        )

    def build_variantes(self, newsolutions):
        variantes = set()
        for sol in newsolutions:
            variante = []
            for key, var_name in self.rewritten.items():
                var = self.select.defined_vars[var_name]
                # skip variable which are only in a NOT EXISTS
                if len(var.stinfo["relations"]) == 1 and isinstance(
                    var.scope.parent, n.Not
                ):
                    continue
                # skip variable whose type is already explicitly specified
                if var.stinfo["typerel"]:
                    continue
                variante.append((key, sol[var_name]))
            if variante:
                variantes.add(tuple(variante))

        if variantes:
            # rebuild variantes as dict
            variantes = [dict(v) for v in variantes]
            # remove variable which have always the same type
            for key in self.rewritten:
                it = iter(variantes)
                etype = next(it)[key]
                for variante in it:
                    if variante[key] != etype:
                        break
                else:
                    for variante in variantes:
                        del variante[key]

        return variantes

    def _cleanup_inserted(self, node):
        # cleanup inserted variable references
        removed = set()
        for vref in node.iget_nodes(n.VariableRef):
            vref.unregister_reference()
            if not vref.variable.stinfo["references"]:
                # no more references, undefine the variable
                del self.select.defined_vars[vref.name]
                removed.add(vref.name)
        for key, newvar in list(self.rewritten.items()):
            if newvar in removed:
                del self.rewritten[key]

    def _may_be_shared_with(self, sniprel, target):
        """if the snippet relation can be skipped to use a relation from the
        original query, return that relation node
        """
        if sniprel.neged(strict=True):
            return None  # no way
        rschema = self.schema.relation_schema_for(sniprel.r_type)
        stmt = self.current_statement()
        for vi in self.varinfos:
            try:
                if target == "object":
                    orels = vi["lhs_rels"][sniprel.r_type]
                    cardindex = 0
                    ttypes_func = rschema.objects
                    rdef = rschema.rdef
                else:  # target == 'subject':
                    orels = vi["rhs_rels"][sniprel.r_type]
                    cardindex = 1
                    ttypes_func = rschema.subjects

                    def rdef(x, y):
                        return rschema.relation_definition(y, x)

            except KeyError:
                # may be raised by vi['xhs_rels'][sniprel.r_type]
                continue
            # if cardinality isn't in '?1', we can't ignore the snippet relation
            # and use variable from the original query
            if _has_multiple_cardinality(
                vi["stinfo"]["possibletypes"], rdef, ttypes_func, cardindex
            ):
                continue
            orel = _compatible_relation(orels, stmt, sniprel)
            if orel is not None:
                return orel
        return None

    def _use_orig_term(self, snippet_varname, term):
        key = (self.current_expr, self.varmap, snippet_varname)
        if key in self.rewritten:
            stmt = self.current_statement()
            insertedvar = stmt.defined_vars.pop(self.rewritten[key])
            for inserted_vref in insertedvar.references():
                inserted_vref.parent.replace(inserted_vref, term.copy(stmt))
        self.rewritten[key] = term.name

    def _get_varname_or_term(self, vname):
        stmt = self.current_statement()
        if vname == "U":
            stmt = self.select
            if self.u_varname is None:
                self.u_varname = stmt.allocate_varname()
                # generate an identifier for the substitution
                argname = stmt.allocate_varname()
                while argname in self.kwargs:
                    argname = stmt.allocate_varname()
                # insert "U eid %(u)s"
                stmt.add_constant_restriction(
                    stmt.get_variable(self.u_varname), "eid", argname, "Substitute"
                )
                self.kwargs[argname] = self.session.user.eid
            return self.u_varname
        key = (self.current_expr, self.varmap, vname)
        try:
            return self.rewritten[key]
        except KeyError:
            self.rewritten[key] = newvname = stmt.allocate_varname()
            return newvname

    # visitor methods ##########################################################

    def _visit_binary(self, node, cls):
        newnode = cls()
        for c in node.children:
            new = c.accept(self)
            if new is None:
                continue
            newnode.append(new)
        if len(newnode.children) == 0:
            return None
        if len(newnode.children) == 1:
            return newnode.children[0]
        return newnode

    def _visit_unary(self, node, cls):
        newc = node.children[0].accept(self)
        if newc is None:
            return None
        newnode = cls()
        newnode.append(newc)
        return newnode

    def visit_and(self, node):
        return self._visit_binary(node, n.And)

    def visit_or(self, node):
        return self._visit_binary(node, n.Or)

    def visit_not(self, node):
        return self._visit_unary(node, n.Not)

    def visit_exists(self, node):
        return self._visit_unary(node, n.Exists)

    def keep_var(self, varname):
        if varname in "SO":
            return varname in self.existingvars
        if varname == "U":
            return True
        vargraph = self.current_expr.vargraph
        for existingvar in self.existingvars:
            if varname not in vargraph or has_path(vargraph, varname, existingvar):
                return True
        # no path from this variable to an existing variable
        return False

    def visit_relation(self, node):
        lhs, rhs = node.get_variable_parts()
        # remove relations where an unexistant variable and or a variable linked
        # to an unexistant variable is used.
        if self.existingvars:
            if not self.keep_var(lhs.name):
                return
        if node.r_type in (
            "has_add_permission",
            "has_update_permission",
            "has_delete_permission",
            "has_read_permission",
        ):
            assert lhs.name == "U"
            action = node.r_type.split("_")[1]
            key = (self.current_expr, self.varmap, rhs.name)
            self.pending_keys.append((key, action))
            return
        if isinstance(rhs, n.VariableRef):
            if self.existingvars and not self.keep_var(rhs.name):
                return
            if lhs.name in self.revvarmap and rhs.name != "U":
                orel = self._may_be_shared_with(node, "object")
                if orel is not None:
                    self._use_orig_term(rhs.name, orel.children[1].children[0])
                    return
            elif rhs.name in self.revvarmap and lhs.name != "U":
                orel = self._may_be_shared_with(node, "subject")
                if orel is not None:
                    self._use_orig_term(lhs.name, orel.children[0])
                    return
        rel = n.Relation(node.r_type, node.optional)
        for c in node.children:
            rel.append(c.accept(self))
        return rel

    def visit_comparison(self, node):
        cmp_ = n.Comparison(node.operator)
        for c in node.children:
            cmp_.append(c.accept(self))
        return cmp_

    def visit_mathexpression(self, node):
        cmp_ = n.MathExpression(node.operator)
        for c in node.children:
            cmp_.append(c.accept(self))
        return cmp_

    def visit_function(self, node):
        """generate filter name for a function"""
        function_ = n.Function(node.name)
        for c in node.children:
            function_.append(c.accept(self))
        return function_

    def visit_constant(self, node):
        """generate filter name for a constant"""
        return n.Constant(node.value, node.type)

    def visit_variableref(self, node):
        """get the sql name for a variable reference"""
        stmt = self.current_statement()
        if node.name in self.revvarmap:
            selectvar, index = self.revvarmap[node.name]
            vi = self.varinfos[index]
            if vi.get("const") is not None:
                return n.Constant(vi["const"], "Int")
            return n.VariableRef(stmt.get_variable(selectvar))
        vname_or_term = self._get_varname_or_term(node.name)
        if isinstance(vname_or_term, str):
            return n.VariableRef(stmt.get_variable(vname_or_term))
        # shared term
        return vname_or_term.copy(stmt)

    def current_statement(self):
        if self._insert_scope is None:
            return self.select
        return self._insert_scope.stmt


class RQLRelationRewriter(RQLRewriter):
    """Insert some rql snippets into another rql syntax tree, replacing computed
    relations by their associated rule.

    This class *isn't thread safe*.
    """

    def rewrite(self, union, kwargs=None):
        self.kwargs = kwargs
        self.removing_ambiguity = False
        self.existingvars = None
        self.pending_keys = None
        rules = self.schema.rules_rqlexpr_mapping
        for relation in union.iget_nodes(n.Relation):
            if relation.r_type in rules:
                self.select = relation.stmt
                self.solutions = self.select.solutions[:]
                self.current_expr = rules[relation.r_type]
                self._insert_scope = relation.scope
                self.rewritten = {}
                lhs, rhs = relation.get_variable_parts()
                varmap = {lhs.name: "S", rhs.name: "O"}
                self.init_from_varmap(tuple(sorted(varmap.items())))
                self.insert_snippet(varmap, self.current_expr.snippet_rqlst)
                self.select.remove_node(relation)

    def _subquery_variable(self, selectvar):
        return self.select.aliases[selectvar].stinfo

    def _inserted_root(self, new):
        return new



# copyright 2003-2014 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""core hooks registering some maintainance tasks as server startup time"""


from datetime import timedelta, datetime

from cubicweb.server import hook, create_user


class TransactionsCleanupStartupHook(hook.Hook):
    """start task to cleanup transaction data"""

    __regid__ = "cw.looping-tasks.transactions-cleanup"
    events = ("server_startup",)

    def __call__(self):
        if not self.repo.has_scheduler():
            return
        # XXX use named args and inner functions to avoid referencing globals
        # which may cause reloading pb
        lifetime = timedelta(days=self.repo.config["keep-transaction-lifetime"])

        def cleanup_old_transactions(repo=self.repo, lifetime=lifetime):
            mindate = datetime.utcnow() - lifetime
            with repo.internal_cnx() as cnx:
                cnx.system_sql(
                    "DELETE FROM transactions WHERE tx_time < %(time)s",
                    {"time": mindate},
                )
                cnx.commit()

        if self.repo.config["undo-enabled"]:
            self.repo.looping_task(60 * 60 * 24, cleanup_old_transactions, self.repo)


class UpdateFeedsStartupHook(hook.Hook):
    """start task to update datafeed based sources"""

    __regid__ = "cw.looping-tasks.update-feeds"
    events = ("server_startup",)

    def __call__(self):
        if not self.repo.has_scheduler():
            return

        def update_feeds(repo):
            # take a list to avoid iterating on a dictionary whose size may
            # change
            for uri, source in repo.sources_by_uri.items():
                if (
                    uri == "system"
                    or not repo.config.source_enabled(source)
                    or not source.config["synchronize"]
                ):
                    continue
                with repo.internal_cnx() as cnx:
                    try:
                        source.pull_data(cnx)
                    except Exception:
                        cnx.exception("while trying to update feed %s", source)

        self.repo.looping_task(60, update_feeds, self.repo)


class DataImportsCleanupStartupHook(hook.Hook):
    """start task to cleanup old data imports (ie datafeed import logs)"""

    __regid__ = "cw.looping-tasks.dataimports-cleanup"
    events = ("server_startup",)

    def __call__(self):
        if not self.repo.has_scheduler():
            return

        def expire_dataimports(repo=self.repo):
            for uri, source in repo.sources_by_uri.items():
                if uri == "system" or not repo.config.source_enabled(source):
                    continue
                with repo.internal_cnx() as cnx:
                    mindate = datetime.utcnow() - timedelta(
                        seconds=source.config["logs-lifetime"]
                    )
                    cnx.execute(
                        "DELETE CWDataImport X WHERE X start_timestamp < %(time)s",
                        {"time": mindate},
                    )
                    cnx.commit()

        self.repo.looping_task(60 * 60 * 24, expire_dataimports, self.repo)


class CreateAnonymousUserIfNeededStartupHook(hook.Hook):
    """Create the anonymous user if needed"""

    __regid__ = "cw.create-anonymous-user"
    events = ("server_startup",)

    def __call__(self):
        if self.repo.config.creating:
            return

        anon_login, anon_password = self.repo.config.anonymous_user()
        if not anon_login:
            return

        with self.repo.internal_cnx() as cnx:
            if not cnx.find("CWUser", login=anon_login):
                create_user(cnx, anon_login, anon_password, "guests")
                self.debug("Create an anonymous user with login %s", anon_login)
            cnx.commit()



# copyright 2003-2014 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.

from yams import ValidationError
from cubicweb import _
from cubicweb.devtools.testlib import CubicWebTC


def add_wf(shell, etype, name=None, default=False):
    if name is None:
        name = etype
    return shell.add_workflow(name, etype, default=default, ensure_workflowable=False)


def parse_hist(wfhist):
    return [
        (
            ti.previous_state.name,
            ti.new_state.name,
            ti.transition and ti.transition.name,
            ti.comment,
        )
        for ti in wfhist
    ]


class WorkflowBuildingTC(CubicWebTC):
    def test_wf_construction(self):
        with self.admin_access.shell() as shell:
            wf = add_wf(shell, "Company")
            foo = wf.add_state("foo", initial=True)
            bar = wf.add_state("bar")
            self.assertEqual(wf.state_by_name("bar").eid, bar.eid)
            self.assertEqual(wf.state_by_name("barrr"), None)
            baz = wf.add_transition("baz", (foo,), bar, ("managers",))
            self.assertEqual(wf.transition_by_name("baz").eid, baz.eid)
            self.assertEqual(len(baz.require_group), 1)
            self.assertEqual(baz.require_group[0].name, "managers")

    def test_duplicated_state(self):
        with self.admin_access.shell() as shell:
            wf = add_wf(shell, "Company")
            wf.add_state("foo", initial=True)
            shell.commit()
            with self.assertRaises(ValidationError) as cm:
                wf.add_state("foo")
            self.assertEqual(
                {
                    "name": "%(KEY-rtype)s is part of violated unicity constraint",
                    "state_of": "%(KEY-rtype)s is part of violated unicity constraint",
                    "": "some relations violate a unicity constraint",
                },
                cm.exception.errors,
            )
            shell.rollback()
            # no pb if not in the same workflow
            wf2 = add_wf(shell, "Company")
            wf2.add_state("foo", initial=True)
            shell.commit()
            # gnark gnark
            bar = wf.add_state("bar")
            shell.commit()
            with self.assertRaises(ValidationError) as cm:
                bar.cw_set(name="foo")
            shell.rollback()
            self.assertEqual(
                {
                    "name": "%(KEY-rtype)s is part of violated unicity constraint",
                    "state_of": "%(KEY-rtype)s is part of violated unicity constraint",
                    "": "some relations violate a unicity constraint",
                },
                cm.exception.errors,
            )

    def test_duplicated_transition(self):
        with self.admin_access.shell() as shell:
            wf = add_wf(shell, "Company")
            foo = wf.add_state("foo", initial=True)
            bar = wf.add_state("bar")
            wf.add_transition("baz", (foo,), bar, ("managers",))
            with self.assertRaises(ValidationError) as cm:
                wf.add_transition("baz", (bar,), foo)
            self.assertEqual(
                {
                    "name": "%(KEY-rtype)s is part of violated unicity constraint",
                    "transition_of": (
                        "%(KEY-rtype)s is part of violated unicity constraint"
                    ),
                    "": "some relations violate a unicity constraint",
                },
                cm.exception.errors,
            )
            shell.rollback()
            # no pb if not in the same workflow
            wf2 = add_wf(shell, "Company")
            foo = wf2.add_state("foo", initial=True)
            bar = wf2.add_state("bar")
            wf2.add_transition("baz", (foo,), bar, ("managers",))
            shell.commit()
            # gnark gnark
            biz = wf2.add_transition("biz", (bar,), foo)
            shell.commit()
            with self.assertRaises(ValidationError) as cm:
                biz.cw_set(name="baz")
            shell.rollback()
            self.assertEqual(
                {
                    "name": "%(KEY-rtype)s is part of violated unicity constraint",
                    "transition_of": (
                        "%(KEY-rtype)s is part of violated unicity constraint"
                    ),
                    "": "some relations violate a unicity constraint",
                },
                cm.exception.errors,
            )


class WorkflowTC(CubicWebTC):
    def setup_database(self):
        rschema = self.schema["in_state"]
        for rdef in rschema.relation_definitions.values():
            self.assertEqual(rdef.cardinality, "1*")
        with self.admin_access.client_cnx() as cnx:
            self.member_eid = self.create_user(cnx, "member").eid
            cnx.commit()

    def test_workflow_base(self):
        with self.admin_access.cnx() as cnx:
            e = self.create_user(cnx, "toto")
            iworkflowable = e.cw_adapt_to("IWorkflowable")
            self.assertEqual(iworkflowable.state, "activated")
            iworkflowable.change_state("deactivated", "deactivate 1")
            cnx.commit()
            iworkflowable.change_state("activated", "activate 1")
            cnx.commit()
            iworkflowable.change_state("deactivated", "deactivate 2")
            cnx.commit()
            e.cw_clear_relation_cache("wf_info_for", "object")
            self.assertEqual(
                [tr.comment for tr in e.reverse_wf_info_for],
                ["deactivate 1", "activate 1", "deactivate 2"],
            )
            self.assertEqual(iworkflowable.latest_trinfo().comment, "deactivate 2")

    def test_possible_transitions(self):
        with self.admin_access.cnx() as cnx:
            user = cnx.execute("CWUser X").get_entity(0, 0)
            iworkflowable = user.cw_adapt_to("IWorkflowable")
            trs = list(iworkflowable.possible_transitions())
            self.assertEqual(len(trs), 1)
            self.assertEqual(trs[0].name, "deactivate")
            self.assertEqual(trs[0].destination(None).name, "deactivated")
        # test a std user get no possible transition
        with self.new_access("member").cnx() as cnx:
            # fetch the entity using the new session
            trs = list(cnx.user.cw_adapt_to("IWorkflowable").possible_transitions())
            self.assertEqual(len(trs), 0)

    def _test_manager_deactivate(self, user):
        iworkflowable = user.cw_adapt_to("IWorkflowable")
        user.cw_clear_relation_cache("in_state", "subject")
        self.assertEqual(len(user.in_state), 1)
        self.assertEqual(iworkflowable.state, "deactivated")
        trinfo = iworkflowable.latest_trinfo()
        self.assertEqual(trinfo.previous_state.name, "activated")
        self.assertEqual(trinfo.new_state.name, "deactivated")
        self.assertEqual(trinfo.comment, "deactivate user")
        self.assertEqual(trinfo.comment_format, "text/plain")
        return trinfo

    def test_change_state(self):
        with self.admin_access.client_cnx() as cnx:
            user = cnx.user
            iworkflowable = user.cw_adapt_to("IWorkflowable")
            iworkflowable.change_state("deactivated", comment="deactivate user")
            trinfo = self._test_manager_deactivate(user)
            self.assertEqual(trinfo.transition, None)

    def test_set_in_state_bad_wf(self):
        with self.admin_access.shell() as shell:
            wf = add_wf(shell, "CWUser")
            s = wf.add_state("foo", initial=True)
            shell.commit()
        with self.admin_access.repo_cnx() as cnx:
            with cnx.security_enabled(write=False):
                with self.assertRaises(ValidationError) as cm:
                    cnx.execute(
                        "SET X in_state S WHERE X eid %(x)s, S eid %(s)s",
                        {"x": cnx.user.eid, "s": s.eid},
                    )
                self.assertEqual(
                    cm.exception.errors,
                    {
                        "in_state-subject": (
                            "state doesn't belong to entity's workflow. "
                            "You may want to set a custom workflow for this entity first."
                        )
                    },
                )

    def test_fire_transition(self):
        with self.admin_access.client_cnx() as cnx:
            user = cnx.user
            iworkflowable = user.cw_adapt_to("IWorkflowable")
            iworkflowable.fire_transition("deactivate", comment="deactivate user")
            user.cw_clear_all_caches()
            self.assertEqual(iworkflowable.state, "deactivated")
            self._test_manager_deactivate(user)
            trinfo = self._test_manager_deactivate(user)
            self.assertEqual(trinfo.transition.name, "deactivate")

    def test_goback_transition(self):
        with self.admin_access.cnx() as cnx:
            wf = cnx.user.cw_adapt_to("IWorkflowable").current_workflow
            asleep = wf.add_state("asleep")
            wf.add_transition(
                "rest",
                (wf.state_by_name("activated"), wf.state_by_name("deactivated")),
                asleep,
            )
            wf.add_transition("wake up", asleep)
            user = self.create_user(cnx, "stduser")
            iworkflowable = user.cw_adapt_to("IWorkflowable")
            iworkflowable.fire_transition("rest")
            cnx.commit()
            iworkflowable.fire_transition("wake up")
            cnx.commit()
            self.assertEqual(iworkflowable.state, "activated")
            iworkflowable.fire_transition("deactivate")
            cnx.commit()
            iworkflowable.fire_transition("rest")
            cnx.commit()
            iworkflowable.fire_transition("wake up")
            cnx.commit()
            user.cw_clear_all_caches()
            self.assertEqual(iworkflowable.state, "deactivated")

    # XXX test managers can change state without matching transition

    def _test_stduser_deactivate(self):
        with self.admin_access.repo_cnx() as cnx:
            self.create_user(cnx, "tutu")
        with self.new_access("tutu").cnx() as cnx:
            iworkflowable = cnx.entity_from_eid(self.member_eid).cw_adapt_to(
                "IWorkflowable"
            )
            with self.assertRaises(ValidationError) as cm:
                iworkflowable.fire_transition("deactivate")
            self.assertEqual(
                cm.exception.errors,
                {"by_transition-subject": "transition may not be fired"},
            )
        with self.new_access("member").cnx() as cnx:
            iworkflowable = cnx.entity_from_eid(self.member_eid).cw_adapt_to(
                "IWorkflowable"
            )
            iworkflowable.fire_transition("deactivate")
            cnx.commit()
            with self.assertRaises(ValidationError) as cm:
                iworkflowable.fire_transition("activate")
            self.assertEqual(
                cm.exception.errors,
                {"by_transition-subject": "transition may not be fired"},
            )

    def test_fire_transition_owned_by(self):
        with self.admin_access.repo_cnx() as cnx:
            cnx.execute(
                'INSERT RQLExpression X: X exprtype "ERQLExpression", '
                'X expression "X owned_by U", T condition X '
                'WHERE T name "deactivate"'
            )
            cnx.commit()
        self._test_stduser_deactivate()

    def test_fire_transition_has_update_perm(self):
        with self.admin_access.repo_cnx() as cnx:
            cnx.execute(
                'INSERT RQLExpression X: X exprtype "ERQLExpression", '
                'X expression "U has_update_permission X", T condition X '
                'WHERE T name "deactivate"'
            )
            cnx.commit()
        self._test_stduser_deactivate()

    def test_swf_base(self):
        """subworkflow

        +-----------+  tr1   +-----------+
        | swfstate1 | ------>| swfstate2 |
        +-----------+        +-----------+
                  |  tr2  +-----------+
                  `------>| swfstate3 |
                          +-----------+

        main workflow

        +--------+  swftr1             +--------+
        | state1 | -------[swfstate2]->| state2 |
        +--------+     |               +--------+
                       |               +--------+
                       `-[swfstate3]-->| state3 |
                                       +--------+
        """
        # sub-workflow
        with self.admin_access.shell() as shell:
            swf = add_wf(shell, "CWGroup", name="subworkflow")
            swfstate1 = swf.add_state("swfstate1", initial=True)
            swfstate2 = swf.add_state("swfstate2")
            swfstate3 = swf.add_state("swfstate3")
            swf.add_transition("tr1", (swfstate1,), swfstate2)
            swf.add_transition("tr2", (swfstate1,), swfstate3)
            # main workflow
            mwf = add_wf(shell, "CWGroup", name="main workflow", default=True)
            state1 = mwf.add_state("state1", initial=True)
            state2 = mwf.add_state("state2")
            state3 = mwf.add_state("state3")
            swftr1 = mwf.add_wftransition(
                "swftr1", swf, state1, [(swfstate2, state2), (swfstate3, state3)]
            )
            swf.cw_clear_all_caches()
            self.assertEqual(swftr1.destination(None).eid, swfstate1.eid)
        # workflows built, begin test
        with self.admin_access.cnx() as cnx:
            group = cnx.create_entity("CWGroup", name="grp1")
            cnx.commit()
            iworkflowable = group.cw_adapt_to("IWorkflowable")
            self.assertEqual(iworkflowable.current_state.eid, state1.eid)
            self.assertEqual(iworkflowable.current_workflow.eid, mwf.eid)
            self.assertEqual(iworkflowable.main_workflow.eid, mwf.eid)
            self.assertEqual(iworkflowable.subworkflow_input_transition(), None)
            iworkflowable.fire_transition("swftr1", "go")
            cnx.commit()
            group.cw_clear_all_caches()
            self.assertEqual(iworkflowable.current_state.eid, swfstate1.eid)
            self.assertEqual(iworkflowable.current_workflow.eid, swf.eid)
            self.assertEqual(iworkflowable.main_workflow.eid, mwf.eid)
            self.assertEqual(
                iworkflowable.subworkflow_input_transition().eid, swftr1.eid
            )
            iworkflowable.fire_transition("tr1", "go")
            cnx.commit()
            group.cw_clear_all_caches()
            self.assertEqual(iworkflowable.current_state.eid, state2.eid)
            self.assertEqual(iworkflowable.current_workflow.eid, mwf.eid)
            self.assertEqual(iworkflowable.main_workflow.eid, mwf.eid)
            self.assertEqual(iworkflowable.subworkflow_input_transition(), None)
            # force back to swfstate1 is impossible since we can't any more find
            # subworkflow input transition
            with self.assertRaises(ValidationError) as cm:
                iworkflowable.change_state(swfstate1, "gadget")
            self.assertEqual(
                cm.exception.errors,
                {"to_state-subject": "state doesn't belong to entity's workflow"},
            )
            cnx.rollback()
            # force back to state1
            iworkflowable.change_state("state1", "gadget")
            iworkflowable.fire_transition("swftr1", "au")
            group.cw_clear_all_caches()
            iworkflowable.fire_transition("tr2", "chapeau")
            cnx.commit()
            group.cw_clear_all_caches()
            self.assertEqual(iworkflowable.current_state.eid, state3.eid)
            self.assertEqual(iworkflowable.current_workflow.eid, mwf.eid)
            self.assertEqual(iworkflowable.main_workflow.eid, mwf.eid)
            self.assertListEqual(
                parse_hist(iworkflowable.workflow_history),
                [
                    ("state1", "swfstate1", "swftr1", "go"),
                    ("swfstate1", "swfstate2", "tr1", "go"),
                    (
                        "swfstate2",
                        "state2",
                        "swftr1",
                        "exiting from subworkflow subworkflow",
                    ),
                    ("state2", "state1", None, "gadget"),
                    ("state1", "swfstate1", "swftr1", "au"),
                    ("swfstate1", "swfstate3", "tr2", "chapeau"),
                    (
                        "swfstate3",
                        "state3",
                        "swftr1",
                        "exiting from subworkflow subworkflow",
                    ),
                ],
            )

    def test_swf_exit_consistency(self):
        with self.admin_access.shell() as shell:
            # sub-workflow
            swf = add_wf(shell, "CWGroup", name="subworkflow")
            swfstate1 = swf.add_state("swfstate1", initial=True)
            swfstate2 = swf.add_state("swfstate2")
            swf.add_transition("tr1", (swfstate1,), swfstate2)
            # main workflow
            mwf = add_wf(shell, "CWGroup", name="main workflow", default=True)
            state1 = mwf.add_state("state1", initial=True)
            state2 = mwf.add_state("state2")
            state3 = mwf.add_state("state3")
            mwf.add_wftransition(
                "swftr1", swf, state1, [(swfstate2, state2), (swfstate2, state3)]
            )
            with self.assertRaises(ValidationError) as cm:
                shell.commit()
            self.assertEqual(
                cm.exception.errors,
                {
                    "subworkflow_exit-subject": (
                        "can't have multiple exits on the same state"
                    )
                },
            )

    def test_swf_fire_in_a_row(self):
        with self.admin_access.shell() as shell:
            # sub-workflow
            subwf = add_wf(shell, "CWGroup", name="subworkflow")
            xsigning = subwf.add_state("xsigning", initial=True)
            xaborted = subwf.add_state("xaborted")
            xsigned = subwf.add_state("xsigned")
            subwf.add_transition("xabort", (xsigning,), xaborted)
            subwf.add_transition("xsign", (xsigning,), xsigning)
            subwf.add_transition("xcomplete", (xsigning,), xsigned, type="auto")
            # main workflow
            twf = add_wf(shell, "CWGroup", name="mainwf", default=True)
            created = twf.add_state(_("created"), initial=True)
            identified = twf.add_state(_("identified"))
            released = twf.add_state(_("released"))
            closed = twf.add_state(_("closed"))
            twf.add_wftransition(
                _("identify"),
                subwf,
                (created,),
                [(xsigned, identified), (xaborted, created)],
            )
            twf.add_wftransition(
                _("release"),
                subwf,
                (identified,),
                [(xsigned, released), (xaborted, identified)],
            )
            twf.add_wftransition(
                _("close"),
                subwf,
                (released,),
                [(xsigned, closed), (xaborted, released)],
            )
            shell.commit()
        with self.admin_access.repo_cnx() as cnx:
            group = cnx.create_entity("CWGroup", name="grp1")
            cnx.commit()
            iworkflowable = group.cw_adapt_to("IWorkflowable")
            for trans in ("identify", "release", "close"):
                iworkflowable.fire_transition(trans)
                cnx.commit()

    def test_swf_magic_tr(self):
        with self.admin_access.shell() as shell:
            # sub-workflow
            subwf = add_wf(shell, "CWGroup", name="subworkflow")
            xsigning = subwf.add_state("xsigning", initial=True)
            xaborted = subwf.add_state("xaborted")
            xsigned = subwf.add_state("xsigned")
            subwf.add_transition("xabort", (xsigning,), xaborted)
            subwf.add_transition("xsign", (xsigning,), xsigned)
            # main workflow
            twf = add_wf(shell, "CWGroup", name="mainwf", default=True)
            created = twf.add_state(_("created"), initial=True)
            identified = twf.add_state(_("identified"))
            twf.add_state(_("released"))
            twf.add_wftransition(
                _("identify"), subwf, created, [(xaborted, None), (xsigned, identified)]
            )
            twf.add_wftransition(_("release"), subwf, identified, [(xaborted, None)])
            shell.commit()
        with self.admin_access.cnx() as cnx:
            group = cnx.create_entity("CWGroup", name="grp1")
            cnx.commit()
            iworkflowable = group.cw_adapt_to("IWorkflowable")
            for trans, nextstate in (
                ("identify", "xsigning"),
                ("xabort", "created"),
                ("identify", "xsigning"),
                ("xsign", "identified"),
                ("release", "xsigning"),
                ("xabort", "identified"),
            ):
                iworkflowable.fire_transition(trans)
                cnx.commit()
                group.cw_clear_all_caches()
                self.assertEqual(iworkflowable.state, nextstate)

    def test_replace_state(self):
        with self.admin_access.shell() as shell:
            wf = add_wf(shell, "CWGroup", name="groupwf", default=True)
            s_new = wf.add_state("new", initial=True)
            s_state1 = wf.add_state("state1")
            wf.add_transition("tr", (s_new,), s_state1)
            shell.commit()

        with self.admin_access.repo_cnx() as cnx:
            group = cnx.create_entity("CWGroup", name="grp1")
            cnx.commit()

            iwf = group.cw_adapt_to("IWorkflowable")
            iwf.fire_transition("tr")
            cnx.commit()
            group.cw_clear_all_caches()

            wf = cnx.entity_from_eid(wf.eid)
            wf.add_state("state2")
            with cnx.security_enabled(write=False):
                wf.replace_state("state1", "state2")
            cnx.commit()

            self.assertEqual(iwf.state, "state2")
            self.assertEqual(iwf.latest_trinfo().to_state[0].name, "state2")


class CustomWorkflowTC(CubicWebTC):
    def setup_database(self):
        with self.admin_access.repo_cnx() as cnx:
            self.member_eid = self.create_user(cnx, "member").eid

    def test_custom_wf_replace_state_no_history(self):
        """member in inital state with no previous history, state is simply
        redirected when changing workflow
        """
        with self.admin_access.shell() as shell:
            wf = add_wf(shell, "CWUser")
            wf.add_state("asleep", initial=True)
        with self.admin_access.cnx() as cnx:
            cnx.execute(
                "SET X custom_workflow WF WHERE X eid %(x)s, WF eid %(wf)s",
                {"wf": wf.eid, "x": self.member_eid},
            )
            member = cnx.entity_from_eid(self.member_eid)
            iworkflowable = member.cw_adapt_to("IWorkflowable")
            self.assertEqual(
                iworkflowable.state, "activated"
            )  # no change before commit
            cnx.commit()
            member.cw_clear_all_caches()
            self.assertEqual(iworkflowable.current_workflow.eid, wf.eid)
            self.assertEqual(iworkflowable.state, "asleep")
            self.assertEqual(iworkflowable.workflow_history, ())

    def test_custom_wf_replace_state_keep_history(self):
        """member in inital state with some history, state is redirected and
        state change is recorded to history
        """
        with self.admin_access.cnx() as cnx:
            member = cnx.entity_from_eid(self.member_eid)
            iworkflowable = member.cw_adapt_to("IWorkflowable")
            iworkflowable.fire_transition("deactivate")
            iworkflowable.fire_transition("activate")
            cnx.commit()
        with self.admin_access.shell() as shell:
            wf = add_wf(shell, "CWUser")
            wf.add_state("asleep", initial=True)
            shell.rqlexec(
                "SET X custom_workflow WF WHERE X eid %(x)s, WF eid %(wf)s",
                {"wf": wf.eid, "x": self.member_eid},
            )
        with self.admin_access.cnx() as cnx:
            member = cnx.entity_from_eid(self.member_eid)
            iworkflowable = member.cw_adapt_to("IWorkflowable")
            self.assertEqual(iworkflowable.current_workflow.eid, wf.eid)
            self.assertEqual(iworkflowable.state, "asleep")
            self.assertEqual(
                parse_hist(iworkflowable.workflow_history),
                [
                    ("activated", "deactivated", "deactivate", None),
                    ("deactivated", "activated", "activate", None),
                    ("activated", "asleep", None, 'workflow changed to "CWUser"'),
                ],
            )

    def test_custom_wf_no_initial_state(self):
        """try to set a custom workflow which has no initial state"""
        with self.admin_access.shell() as shell:
            wf = add_wf(shell, "CWUser")
            wf.add_state("asleep")
            shell.rqlexec(
                "SET X custom_workflow WF WHERE X eid %(x)s, WF eid %(wf)s",
                {"wf": wf.eid, "x": self.member_eid},
            )
            with self.assertRaises(ValidationError) as cm:
                shell.commit()
            self.assertEqual(
                cm.exception.errors,
                {"custom_workflow-subject": "workflow has no initial state"},
            )

    def test_custom_wf_bad_etype(self):
        """try to set a custom workflow which doesn't apply to entity type"""
        with self.admin_access.shell() as shell:
            wf = add_wf(shell, "Company")
            wf.add_state("asleep", initial=True)
            shell.rqlexec(
                "SET X custom_workflow WF WHERE X eid %(x)s, WF eid %(wf)s",
                {"wf": wf.eid, "x": self.member_eid},
            )
            with self.assertRaises(ValidationError) as cm:
                shell.commit()
            self.assertEqual(
                cm.exception.errors,
                {"custom_workflow-subject": "workflow isn't a workflow for this type"},
            )

    def test_del_custom_wf(self):
        """member in some state shared by the new workflow, nothing has to be
        done
        """
        with self.admin_access.cnx() as cnx:
            member = cnx.entity_from_eid(self.member_eid)
            iworkflowable = member.cw_adapt_to("IWorkflowable")
            iworkflowable.fire_transition("deactivate")
            cnx.commit()
        with self.admin_access.shell() as shell:
            wf = add_wf(shell, "CWUser")
            wf.add_state("asleep", initial=True)
            shell.rqlexec(
                "SET X custom_workflow WF WHERE X eid %(x)s, WF eid %(wf)s",
                {"wf": wf.eid, "x": self.member_eid},
            )
            shell.commit()
        with self.admin_access.cnx() as cnx:
            cnx.execute(
                "DELETE X custom_workflow WF WHERE X eid %(x)s, WF eid %(wf)s",
                {"wf": wf.eid, "x": self.member_eid},
            )
            member = cnx.entity_from_eid(self.member_eid)
            iworkflowable = member.cw_adapt_to("IWorkflowable")
            self.assertEqual(iworkflowable.state, "asleep")  # no change before commit
            cnx.commit()
            member.cw_clear_all_caches()
            self.assertEqual(
                iworkflowable.current_workflow.name, "default user workflow"
            )
            self.assertEqual(iworkflowable.state, "activated")
            self.assertEqual(
                parse_hist(iworkflowable.workflow_history),
                [
                    ("activated", "deactivated", "deactivate", None),
                    ("deactivated", "asleep", None, 'workflow changed to "CWUser"'),
                    (
                        "asleep",
                        "activated",
                        None,
                        'workflow changed to "default user workflow"',
                    ),
                ],
            )


class AutoTransitionTC(CubicWebTC):
    def setup_custom_wf(self):
        with self.admin_access.shell() as shell:
            wf = add_wf(shell, "CWUser")
            asleep = wf.add_state("asleep", initial=True)
            dead = wf.add_state("dead")
            wf.add_transition("rest", asleep, asleep)
            wf.add_transition(
                "sick",
                asleep,
                dead,
                type="auto",
                conditions=({"expr": 'X surname "toto"', "mainvars": "X"},),
            )
        return wf

    def test_auto_transition_fired(self):
        wf = self.setup_custom_wf()
        with self.admin_access.cnx() as cnx:
            user = self.create_user(cnx, "member")
            iworkflowable = user.cw_adapt_to("IWorkflowable")
            cnx.execute(
                "SET X custom_workflow WF WHERE X eid %(x)s, WF eid %(wf)s",
                {"wf": wf.eid, "x": user.eid},
            )
            cnx.commit()
            user.cw_clear_all_caches()
            self.assertEqual(iworkflowable.state, "asleep")
            self.assertEqual(
                [t.name for t in iworkflowable.possible_transitions()], ["rest"]
            )
            iworkflowable.fire_transition("rest")
            cnx.commit()
            user.cw_clear_all_caches()
            self.assertEqual(iworkflowable.state, "asleep")
            self.assertEqual(
                [t.name for t in iworkflowable.possible_transitions()], ["rest"]
            )
            self.assertEqual(
                parse_hist(iworkflowable.workflow_history),
                [("asleep", "asleep", "rest", None)],
            )
            user.cw_set(surname="toto")  # fulfill condition
            cnx.commit()
            iworkflowable.fire_transition("rest")
            cnx.commit()
            user.cw_clear_all_caches()
            self.assertEqual(iworkflowable.state, "dead")
            self.assertEqual(
                parse_hist(iworkflowable.workflow_history),
                [
                    ("asleep", "asleep", "rest", None),
                    ("asleep", "asleep", "rest", None),
                    ("asleep", "dead", "sick", None),
                ],
            )

    def test_auto_transition_custom_initial_state_fired(self):
        wf = self.setup_custom_wf()
        with self.admin_access.cnx() as cnx:
            user = self.create_user(cnx, "member", surname="toto")
            cnx.execute(
                "SET X custom_workflow WF WHERE X eid %(x)s, WF eid %(wf)s",
                {"wf": wf.eid, "x": user.eid},
            )
            cnx.commit()
            user.cw_clear_all_caches()
            iworkflowable = user.cw_adapt_to("IWorkflowable")
            self.assertEqual(iworkflowable.state, "dead")

    def test_auto_transition_initial_state_fired(self):
        with self.admin_access.cnx() as cnx:
            wf = cnx.execute(
                "Any WF WHERE ET default_workflow WF, " "ET name %(et)s",
                {"et": "CWUser"},
            ).get_entity(0, 0)
            dead = wf.add_state("dead")
            wf.add_transition(
                "sick",
                wf.state_by_name("activated"),
                dead,
                type="auto",
                conditions=({"expr": 'X surname "toto"', "mainvars": "X"},),
            )
            cnx.commit()
        with self.admin_access.cnx() as cnx:
            user = self.create_user(cnx, "member", surname="toto")
            cnx.commit()
            iworkflowable = user.cw_adapt_to("IWorkflowable")
            self.assertEqual(iworkflowable.state, "dead")


class WorkflowHooksTC(CubicWebTC):
    def setUp(self):
        CubicWebTC.setUp(self)
        with self.admin_access.cnx() as cnx:
            self.wf = cnx.user.cw_adapt_to("IWorkflowable").current_workflow
            self.s_activated = self.wf.state_by_name("activated").eid
            self.s_deactivated = self.wf.state_by_name("deactivated").eid
            self.s_dummy = self.wf.add_state("dummy").eid
            self.wf.add_transition("dummy", (self.s_deactivated,), self.s_dummy)
            ueid = self.create_user(cnx, "stduser", commit=False).eid
            # test initial state is set
            rset = cnx.execute(
                "Any N WHERE S name N, X in_state S, X eid %(x)s", {"x": ueid}
            )
            self.assertFalse(rset, rset.rows)
            cnx.commit()
            initialstate = cnx.execute(
                "Any N WHERE S name N, X in_state S, X eid %(x)s", {"x": ueid}
            )[0][0]
            self.assertEqual(initialstate, "activated")
            # give access to users group on the user's wf transitions
            # so we can test wf enforcing on euser (managers don't have anymore this
            # enforcement
            cnx.execute(
                "SET X require_group G "
                'WHERE G name "users", X transition_of WF, WF eid %(wf)s',
                {"wf": self.wf.eid},
            )
            cnx.commit()

    # XXX currently, we've to rely on hooks to set initial state, or to use execute
    # def test_initial_state(self):
    #     cnx = self.login('stduser')
    #     cu = cnx.cursor()
    #     self.assertRaises(ValidationError, cu.execute,
    #                       'INSERT CWUser X: X login "badaboum", X upassword %(pwd)s, '
    #                       'X in_state S WHERE S name "deactivated"', {'pwd': 'oops'})
    #     cnx.close()
    #     # though managers can do whatever he want
    #     self.execute('INSERT CWUser X: X login "badaboum", X upassword %(pwd)s, '
    #                  'X in_state S, X in_group G WHERE S name "deactivated", G name "users"', {'pwd': 'oops'})
    #     self.commit()

    # test that the workflow is correctly enforced

    def _cleanup_msg(self, msg):
        """remove the variable part of one specific error message"""
        lmsg = msg.split()
        lmsg.pop(1)
        lmsg.pop()
        return " ".join(lmsg)

    def test_transition_checking1(self):
        with self.new_access("stduser").repo_cnx() as cnx:
            user = cnx.user
            iworkflowable = user.cw_adapt_to("IWorkflowable")
            with self.assertRaises(ValidationError) as cm:
                iworkflowable.fire_transition("activate")
            self.assertEqual(
                self._cleanup_msg(cm.exception.errors["by_transition-subject"]),
                "transition isn't allowed from",
            )

    def test_transition_checking2(self):
        with self.new_access("stduser").repo_cnx() as cnx:
            user = cnx.user
            iworkflowable = user.cw_adapt_to("IWorkflowable")
            with self.assertRaises(ValidationError) as cm:
                iworkflowable.fire_transition("dummy")
            self.assertEqual(
                self._cleanup_msg(cm.exception.errors["by_transition-subject"]),
                "transition isn't allowed from",
            )

    def test_transition_checking3(self):
        with self.new_access("stduser").repo_cnx() as cnx:
            user = cnx.user
            iworkflowable = user.cw_adapt_to("IWorkflowable")
            iworkflowable.fire_transition("deactivate")
            cnx.commit()
            with self.assertRaises(ValidationError) as cm:
                iworkflowable.fire_transition("deactivate")
            self.assertEqual(
                self._cleanup_msg(cm.exception.errors["by_transition-subject"]),
                "transition isn't allowed from",
            )
            cnx.rollback()
            # get back now
            iworkflowable.fire_transition("activate")
            cnx.commit()


if __name__ == "__main__":
    from logilab.common.testlib import unittest_main

    unittest_main()
